import pytest
from numpy import array, nan
from numpy.random import RandomState

from limix.qc import mean_standardize

from .util import (
    assert_dask_array,
    assert_dask_dataframe,
    assert_ndarray_1d,
    assert_ndarray_2d,
    assert_pandas_dataframe,
    assert_pandas_series,
    assert_xarray_dataarray,
)


@pytest.fixture
def data2d():
    X = RandomState(0).randn(3, 2)
    X[0, 0] = nan
    R = array(
        [[nan, -0.117_142_199_8], [-1.0, 1.279_107_161_9], [1.0, -1.161_964_962_2]]
    )
    Rt = array([[nan, 0.0], [-1.0, 1.0], [1.0, -1.0]])

    samples = [f"sample{i}" for i in range(X.shape[0])]
    candidates = [f"snp{i}" for i in range(X.shape[1])]

    return {"X": X, "R": R, "Rt": Rt, "samples": samples, "candidates": candidates}


@pytest.fixture
def data1d():
    X = RandomState(0).randn(3)
    X[0] = nan
    R = array([nan, -1.0, 1.0])
    Rt = array([nan, 0.0, 0.0])

    samples = [f"sample{i}" for i in range(X.shape[0])]
    return {"X": X, "R": R, "Rt": Rt, "samples": samples, "candidates": None}


def test_mean_standardize_ndarray_1d(data1d):
    assert_ndarray_1d(mean_standardize, data1d)


def test_mean_standardize_ndarray_2d(data2d):
    assert_ndarray_2d(mean_standardize, data2d)


def test_mean_standardize_pandas_series(data1d):
    assert_pandas_series(mean_standardize, data1d)


def test_mean_standardize_pandas_dataframe(data2d):
    assert_pandas_dataframe(mean_standardize, data2d)


def test_mean_standardize_dask_array(data2d):
    assert_dask_array(mean_standardize, data2d, test_inplace=False)


def test_mean_standardize_dask_dataframe(data2d):
    assert_dask_dataframe(mean_standardize, data2d)


def test_mean_standardize_xarray_dataarray(data2d):
    assert_xarray_dataarray(mean_standardize, data2d)



import warnings

from .._data import conform_dataset, normalize_likelihood
from .._display import session_block


class VarDec(object):
    """
    Variance decompositon through GLMMs.

    Example
    -------

    .. doctest::

        >>> from limix.vardec import VarDec
        >>> from limix.stats import multivariate_normal as mvn
        >>> from numpy import ones, eye, concatenate, zeros, exp
        >>> from numpy.random import RandomState
        >>>
        >>> random = RandomState(0)
        >>> nsamples = 20
        >>>
        >>> M = random.randn(nsamples, 2)
        >>> M = (M - M.mean(0)) / M.std(0)
        >>> M = concatenate((ones((nsamples, 1)), M), axis=1)
        >>>
        >>> K0 = random.randn(nsamples, 10)
        >>> K0 = K0 @ K0.T
        >>> K0 /= K0.diagonal().mean()
        >>> K0 += eye(nsamples) * 1e-4
        >>>
        >>> K1 = random.randn(nsamples, 10)
        >>> K1 = K1 @ K1.T
        >>> K1 /= K1.diagonal().mean()
        >>> K1 += eye(nsamples) * 1e-4
        >>>
        >>> y = M @ random.randn(3) + mvn(random, zeros(nsamples), K0)
        >>> y += mvn(random, zeros(nsamples), K1)
        >>>
        >>> vardec = VarDec(y, "normal", M)
        >>> vardec.append(K0)
        >>> vardec.append(K1)
        >>> vardec.append_iid()
        >>>
        >>> vardec.fit(verbose=False)
        >>> print(vardec) # doctest: +FLOAT_CMP
        Variance decomposition
        ======================
        <BLANKLINE>
        𝐲 ~ 𝓝(𝙼𝜶, 0.385⋅𝙺 + 1.184⋅𝙺 + 0.000⋅𝙸)
        >>> y = exp((y - y.mean()) / y.std())
        >>> vardec = VarDec(y, "poisson", M)
        >>> vardec.append(K0)
        >>> vardec.append(K1)
        >>> vardec.append_iid()
        >>>
        >>> vardec.fit(verbose=False)
        >>> print(vardec) # doctest: +FLOAT_CMP
        Variance decomposition
        ======================
        <BLANKLINE>
        𝐳 ~ 𝓝(𝙼𝜶, 0.000⋅𝙺 + 0.350⋅𝙺 + 0.000⋅𝙸) for yᵢ ~ Poisson(λᵢ=g(zᵢ)) and g(x)=eˣ
    """

    def __init__(self, y, lik="normal", M=None):
        """
        Constructor.

        Parameters
        ----------
        y : array_like
            Phenotype.
        lik : tuple, "normal", "bernoulli", "probit", "binomial", "poisson"
            Sample likelihood describing the residual distribution.
            Either a tuple or a string specifying the likelihood is required. The
            Normal, Bernoulli, Probit, and Poisson likelihoods can be selected by
            providing a string. Binomial likelihood on the other hand requires a tuple
            because of the number of trials: ``("binomial", array_like)``. Defaults to
            ``"normal"``.
        M : n×c array_like
            Covariates matrix.
        """
        from numpy import asarray
        from glimix_core.mean import LinearMean

        y = asarray(y, float)
        data = conform_dataset(y, M)
        y = data["y"]
        M = data["M"]
        self._y = y
        self._M = M
        self._lik = normalize_likelihood(lik)
        self._mean = LinearMean(asarray(M, float))
        self._covariance = []
        self._glmm = None
        self._fit = False
        self._unnamed = 0

    @property
    def effsizes(self):
        """
        Covariace effect sizes.

        Returns
        -------
        effsizes : ndarray
            Effect sizes.
        """
        if not self._fit:
            self.fit()
        return self._mean.effsizes

    @property
    def covariance(self):
        """
        Get the covariance matrices.

        Returns
        -------
        covariances : list
            Covariance matrices.
        """
        return self._covariance

    def fit(self, verbose=True):
        """
        Fit the model.

        Parameters
        ----------
        verbose : bool, optional
            Set ``False`` to silence it. Defaults to ``True``.
        """
        with session_block("Variance decomposition", disable=not verbose):
            if self._lik[0] == "normal":
                if self._simple_model():
                    self._fit_lmm_simple_model(verbose)
                else:
                    self._fit_lmm(verbose)
            else:
                if self._simple_model():
                    self._fit_glmm_simple_model(verbose)
                else:
                    self._fit_glmm(verbose)

            if verbose:
                print(self)

        self._fit = True

    def lml(self):
        """
        Get the log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        if not self._fit:
            self._glmm.fit()
        return self._glmm.lml()

    def append_iid(self, name="residual"):
        from glimix_core.cov import EyeCov

        c = EyeCov(self._y.shape[0])
        c.name = name
        self._covariance.append(c)

    def append(self, K, name=None):
        from numpy_sugar import is_all_finite
        from numpy import asarray
        from glimix_core.cov import GivenCov

        data = conform_dataset(self._y, K=K)
        K = asarray(data["K"], float)

        if not is_all_finite(K):
            raise ValueError("Covariance-matrix values must be finite.")

        K = K / K.diagonal().mean()
        cov = GivenCov(K)
        if name is None:
            name = "unnamed-{}".format(self._unnamed)
            self._unnamed += 1
        cov.name = name

        self._covariance.append(cov)

    def plot(self):
        import limix
        import seaborn as sns
        from matplotlib.ticker import FormatStrFormatter

        variances = [c.scale for c in self._covariance]
        variances = [(v / sum(variances)) * 100 for v in variances]
        names = [c.name for c in self._covariance]

        ax = sns.barplot(x=names, y=variances)
        ax.yaxis.set_major_formatter(FormatStrFormatter("%.0f%%"))
        ax.set_xlabel("random effects")
        ax.set_ylabel("explained variance")
        ax.set_title("Variance decomposition")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            limix.plot.get_pyplot().tight_layout()
        limix.plot.show()

    def _fit_lmm(self, verbose):
        from glimix_core.cov import SumCov
        from glimix_core.gp import GP
        from numpy import asarray

        y = asarray(self._y, float).ravel()
        gp = GP(y, self._mean, SumCov(self._covariance))
        gp.fit(verbose=verbose)
        self._glmm = gp

    def _fit_glmm(self, verbose):
        from glimix_core.cov import SumCov
        from glimix_core.ggp import ExpFamGP
        from numpy import asarray

        y = asarray(self._y, float).ravel()
        gp = ExpFamGP(y, self._lik, self._mean, SumCov(self._covariance))
        gp.fit(verbose=verbose)
        self._glmm = gp

    def _fit_lmm_simple_model(self, verbose):
        from numpy_sugar.linalg import economic_qs
        from glimix_core.lmm import LMM
        from numpy import asarray

        K = self._get_matrix_simple_model()

        y = asarray(self._y, float).ravel()
        QS = None
        if K is not None:
            QS = economic_qs(K)
        lmm = LMM(y, self._M, QS)
        lmm.fit(verbose=verbose)
        self._set_simple_model_variances(lmm.v0, lmm.v1)
        self._glmm = lmm

    def _fit_glmm_simple_model(self, verbose):
        from numpy_sugar.linalg import economic_qs
        from glimix_core.glmm import GLMMExpFam
        from numpy import asarray

        K = self._get_matrix_simple_model()

        y = asarray(self._y, float).ravel()
        QS = None
        if K is not None:
            QS = economic_qs(K)

        glmm = GLMMExpFam(y, self._lik, self._M, QS)
        glmm.fit(verbose=verbose)

        self._set_simple_model_variances(glmm.v0, glmm.v1)
        self._glmm = glmm

    def _set_simple_model_variances(self, v0, v1):
        from glimix_core.cov import GivenCov, EyeCov

        for c in self._covariance:
            if isinstance(c, GivenCov):
                c.scale = v0
            elif isinstance(c, EyeCov):
                c.scale = v1

    def _get_matrix_simple_model(self):
        from glimix_core.cov import GivenCov

        K = None
        for i in range(len(self._covariance)):
            if isinstance(self._covariance[i], GivenCov):
                self._covariance[i].scale = 1.0
                K = self._covariance[i].value()
                break
        return K

    def _simple_model(self):
        from glimix_core.cov import GivenCov, EyeCov

        if len(self._covariance) > 2:
            return False

        c = self._covariance
        if len(c) == 1 and isinstance(c[0], EyeCov):
            return True

        if isinstance(c[0], GivenCov) and isinstance(c[1], EyeCov):
            return True

        if isinstance(c[1], GivenCov) and isinstance(c[0], EyeCov):
            return True

        return False

    def __repr__(self):
        from glimix_core.cov import GivenCov
        from limix.qtl._result._draw import draw_model, draw_title

        covariance = ""
        for c in self._covariance:
            s = c.scale
            if isinstance(c, GivenCov):
                covariance += f"{s:.3f}⋅𝙺 + "
            else:
                covariance += f"{s:.3f}⋅𝙸 + "
        if len(covariance) > 2:
            covariance = covariance[:-3]

        msg = draw_title("Variance decomposition")
        msg += draw_model(self._lik[0], "𝙼𝜶", covariance)
        msg = msg.rstrip()

        return msg



from ._assert import assert_likelihood


def normalize_likelihood(lik):
    if not isinstance(lik, (tuple, list)):
        lik = (lik,)

    lik_name = lik[0].lower()
    lik = (lik_name,) + lik[1:]
    assert_likelihood(lik_name)
    return lik



import click


class OrderedCommand(click.Command):
    def parse_args(self, ctx, args):
        parser = self.make_parser(ctx)
        opts, args, param_order = parser.parse_args(args=args)

        for param in click.core.iter_params_for_processing(
            param_order, self.get_params(ctx)
        ):
            value, args = param.handle_parse_result(ctx, opts, args)

        if args and not ctx.allow_extra_args and not ctx.resilient_parsing:
            ctx.fail(
                "Got unexpected extra argument%s (%s)"
                % (
                    len(args) != 1 and "s" or "",
                    " ".join(map(click.utils.make_str, args)),
                )
            )

        ctx.args = args
        ctx.param_order = param_order
        return args



from limix._cache import cache
from limix.stats import lrt_pvalues

from ._aligned import Aligned
from ._draw import draw_alt_hyp_table, draw_lrt_table, draw_model, draw_title


class MTScanResult:
    def __init__(self, tests, traits, covariates, candidates, h0, envs0, envs1):
        self._tests = tests
        self._traits = traits
        self._covariates = covariates
        self._candidates = candidates
        self._envs0 = envs0
        self._envs1 = envs1
        self._h0 = h0

    @property
    def stats(self):
        """
        Statistics.
        """
        return self._dataframes["stats"].set_index("test")

    @property
    def effsizes(self):
        """
        Effect sizes.
        """
        return self._dataframes["effsizes"]

    @property
    def h0(self):
        """
        Hypothesis zero.
        """
        return self._h0

    @property
    def _h0_dataframe(self):
        from pandas import DataFrame

        covariates = list(self._covariates)

        h0 = []
        for i, trait in enumerate(self._traits):
            for j, c in enumerate(covariates):
                eff = self._h0["effsizes"][j, i]
                eff_se = self._h0["effsizes_se"][j, i]
                h0.append([trait, "covariate", c, eff, eff_se])

        columns = ["trait", "effect_type", "effect_name", "effsize", "effsize_se"]
        return DataFrame(h0, columns=columns)

    @property
    def _h1_dataframe(self):
        from pandas import DataFrame

        covariates = list(self._covariates)
        envs0 = list(self._envs0)

        h1 = []
        for i, test in enumerate(self._tests):
            candidates = list(self._candidates[test["idx"]])

            effsizes = test["h1"]["covariate_effsizes"]
            effsizes_se = test["h1"]["covariate_effsizes_se"]
            for j, trait in enumerate(self._traits):
                for l, c in enumerate(covariates):
                    eff = effsizes[l, j]
                    eff_se = effsizes_se[l, j]
                    v = [i, str(trait), "covariate", str(c), None, eff, eff_se]
                    h1.append(v)

            effsizes = test["h1"]["candidate_effsizes"]
            effsizes_se = test["h1"]["candidate_effsizes_se"]
            for j, e in enumerate(envs0):
                for l, c in enumerate(candidates):
                    env_name = "env0_" + str(e)
                    eff = effsizes[l, j]
                    eff_se = effsizes_se[l, j]
                    v = [i, str(trait), "candidate", str(c), env_name, eff, eff_se]
                    h1.append(v)

        columns = [
            "test",
            "trait",
            "effect_type",
            "effect_name",
            "env",
            "effsize",
            "effsize_se",
        ]
        return DataFrame(h1, columns=columns)

    @property
    def _h2_dataframe(self):
        from pandas import DataFrame

        envs0 = list(self._envs0)
        envs1 = list(self._envs1)
        covariates = list(self._covariates)

        h2 = []
        for i, test in enumerate(self._tests):
            candidates = list(self._candidates[test["idx"]])

            effsizes = test["h2"]["covariate_effsizes"]
            effsizes_se = test["h2"]["covariate_effsizes_se"]
            for j, trait in enumerate(self._traits):
                for l, c in enumerate(covariates):
                    eff = effsizes[l, j]
                    eff_se = effsizes_se[l, j]
                    v = [i, str(trait), "covariate", str(c), None, eff, eff_se]
                    h2.append(v)

            effsizes = test["h2"]["candidate_effsizes"]
            effsizes_se = test["h2"]["candidate_effsizes_se"]
            off = 0
            for j, e in enumerate(envs0):
                for l, c in enumerate(candidates):
                    env_name = "env0_" + str(e)
                    eff = effsizes[l, off + j]
                    eff_se = effsizes_se[l, off + j]
                    v = [i, str(trait), "candidate", str(c), env_name, eff, eff_se]
                    h2.append(v)

            off = len(envs0)
            for j, e in enumerate(envs1):
                for l, c in enumerate(candidates):
                    env_name = "env1_" + str(e)
                    eff = effsizes[l, off + j]
                    eff_se = effsizes_se[l, off + j]
                    v = [i, str(trait), "candidate", str(c), env_name, eff, eff_se]
                    h2.append(v)

        columns = [
            "test",
            "trait",
            "effect_type",
            "effect_name",
            "env",
            "effsize",
            "effsize_se",
        ]
        return DataFrame(h2, columns=columns)

    @property
    def _stats_dataframe(self):
        from pandas import DataFrame

        if len(self._envs0) == 0:
            return self._stats_dataframe_h2_only

        stats = []
        for i, test in enumerate(self._tests):
            dof10 = test["h1"]["candidate_effsizes"].size
            dof20 = test["h2"]["candidate_effsizes"].size
            dof21 = dof20 - dof10
            stats.append(
                [
                    i,
                    self._h0.lml,
                    test["h1"]["lml"],
                    test["h2"]["lml"],
                    dof10,
                    dof20,
                    dof21,
                    test["h1"]["scale"],
                    test["h2"]["scale"],
                ]
            )

        columns = [
            "test",
            "lml0",
            "lml1",
            "lml2",
            "dof10",
            "dof20",
            "dof21",
            "scale1",
            "scale2",
        ]
        stats = DataFrame(stats, columns=columns)

        stats["pv10"] = lrt_pvalues(stats["lml0"], stats["lml1"], stats["dof10"])
        stats["pv20"] = lrt_pvalues(stats["lml0"], stats["lml2"], stats["dof20"])
        stats["pv21"] = lrt_pvalues(stats["lml1"], stats["lml2"], stats["dof21"])

        return stats

    @property
    def _stats_dataframe_h2_only(self):
        from pandas import DataFrame

        stats = []
        for i, test in enumerate(self._tests):
            dof20 = test["h2"]["candidate_effsizes"].size
            stats.append(
                [i, self._h0.lml, test["h2"]["lml"], dof20, test["h2"]["scale"]]
            )

        columns = ["test", "lml0", "lml2", "dof20", "scale2"]
        stats = DataFrame(stats, columns=columns)

        stats["pv20"] = lrt_pvalues(stats["lml0"], stats["lml2"], stats["dof20"])

        return stats

    @property
    @cache
    def _dataframes(self):
        h2 = self._h2_dataframe
        stats = self._stats_dataframe
        r = {"stats": stats, "effsizes": {"h2": h2}}
        if len(self._envs0) > 0:
            r["effsizes"]["h1"] = self._h1_dataframe
        return r

    def _repr_three_hypothesis(self):
        from numpy import asarray

        traits = self._h0.traits
        lik = self._h0.likelihood
        covariates = self._covariates
        lml = self._h0.lml
        effsizes = asarray(self.h0.effsizes["effsize"], float).ravel()
        effsizes_se = asarray(self.h0.effsizes["effsize_se"], float).ravel()
        stats = self.stats

        df = self.h0.variances
        df = df[df["trait0"] == df["trait1"]]
        diagC0 = df["fore_covariance"]
        diagC1 = df["back_covariance"]

        msg = draw_title("Hypothesis 0")
        msg += draw_model(lik, "(A⊗𝙼)𝛂", "C₀⊗𝙺 + C₁⊗𝙸") + "\n"
        msg += _draw_hyp0_summary(
            traits, covariates, effsizes, effsizes_se, lml, diagC0, diagC1
        )

        msg += draw_title("Hypothesis 1")
        msg += draw_model(lik, "(A⊗𝙼)𝛂 + (A₀⊗G)𝛃₀", "s(C₀⊗𝙺 + C₁⊗𝙸)")
        msg += draw_alt_hyp_table(1, self.stats, self.effsizes)

        msg += draw_title("Hypothesis 2")
        msg += draw_model(lik, "(A⊗𝙼)𝛂 + (A₀⊗G)𝛃₀ + (A₁⊗G)𝛃₁", "s(C₀⊗𝙺 + C₁⊗𝙸)")
        msg += draw_alt_hyp_table(2, self.stats, self.effsizes)

        msg += draw_title("Likelihood-ratio test p-values")
        cols = ["𝓗₀ vs 𝓗₁", "𝓗₀ vs 𝓗₂", "𝓗₁ vs 𝓗₂"]
        msg += draw_lrt_table(cols, ["pv10", "pv20", "pv21"], stats)
        return msg

    def _repr_two_hypothesis(self, alt_hyp):
        from numpy import asarray

        traits = self._h0.traits
        lik = self._h0.likelihood
        covariates = self._covariates
        lml = self._h0.lml
        effsizes = asarray(self.h0.effsizes["effsize"], float).ravel()
        effsizes_se = asarray(self.h0.effsizes["effsize_se"], float).ravel()
        stats = self.stats

        df = self.h0.variances
        df = df[df["trait0"] == df["trait1"]]
        diagC0 = df["fore_covariance"]
        diagC1 = df["back_covariance"]

        covariance = "C₀⊗𝙺 + C₁⊗𝙸"

        msg = draw_title("Hypothesis 0")
        msg += draw_model(lik, "(A⊗𝙼)𝜶", "C₀⊗𝙺 + C₁⊗𝙸") + "\n"
        msg += _draw_hyp0_summary(
            traits, covariates, effsizes, effsizes_se, lml, diagC0, diagC1
        )

        if alt_hyp == 1:
            mean = "(A⊗𝙼)𝜶 + (A₀⊗G)𝛃₀"
            col = "𝓗₀ vs 𝓗₁"
        else:
            mean = "(A⊗𝙼)𝜶 + (A₁⊗G)𝛃₁"
            col = "𝓗₀ vs 𝓗₂"

        msg += draw_title(f"Hypothesis {alt_hyp}")
        msg += draw_model(lik, mean, f"s({covariance})")
        msg += draw_alt_hyp_table(alt_hyp, self.stats, self.effsizes)

        msg += draw_title("Likelihood-ratio test p-values")
        msg += draw_lrt_table([col], [f"pv{alt_hyp}0"], stats)
        return msg

    def __repr__(self):
        if len(self._envs0) == 0:
            return self._repr_two_hypothesis(2)
        return self._repr_three_hypothesis()


def _draw_hyp0_summary(traits, covariates, effsizes, effsizes_se, lml, diagC0, diagC1):
    from numpy import asarray

    diagC0 = asarray(diagC0, float)
    diagC1 = asarray(diagC1, float)
    aligned = Aligned()
    aligned.add_item("traits", traits)
    aligned.add_item("M", covariates)
    aligned.add_item("𝜶", effsizes)
    aligned.add_item("se(𝜶)", effsizes_se)
    aligned.add_item("diag(C₀)", diagC0)
    aligned.add_item("diag(C₁)", diagC1)
    aligned.add_item("lml", lml)
    return aligned.draw() + "\n"



from ._table import Table


def draw_title(title):
    msg = f"{title}\n"
    msg += "=" * len(title) + "\n\n"
    return msg


def draw_model(lik, mean, covariance):
    if lik == "normal":
        var = "𝐲"
    else:
        var = "𝐳"

    msg = f"{var} ~ 𝓝({mean}, {covariance})"
    msg += _lik_formulae(lik)
    return msg


def draw_alt_hyp_table(hyp_num, stats, effsizes):
    cols = ["lml", "cov. effsizes", "cand. effsizes"]
    table = Table(cols, index=_describe_index())
    table.add_column(_describe(stats, f"lml{hyp_num}"))
    df = effsizes[f"h{hyp_num}"]
    table.add_column(_describe(df[df["effect_type"] == "covariate"], "effsize"))
    table.add_column(_describe(df[df["effect_type"] == "candidate"], "effsize"))
    return "\n" + table.draw() + "\n\n"


def draw_lrt_table(test_titles, pv_names, stats):
    table = Table(test_titles, index=_describe_index())

    for name in pv_names:
        pv = stats[name].describe().iloc[1:]
        table.add_column(pv)

    return table.draw()


def _lik_formulae(lik):
    msg = ""

    if lik == "bernoulli":
        msg += f" for yᵢ ~ Bern(μᵢ=g(zᵢ)) and g(x)=1/(1+e⁻ˣ)\n"
    elif lik == "probit":
        msg += f" for yᵢ ~ Bern(μᵢ=g(zᵢ)) and g(x)=Φ(x)\n"
    elif lik == "binomial":
        msg += f" for yᵢ ~ Binom(μᵢ=g(zᵢ), nᵢ) and g(x)=1/(1+e⁻ˣ)\n"
    elif lik == "poisson":
        msg += f" for yᵢ ~ Poisson(λᵢ=g(zᵢ)) and g(x)=eˣ\n"
    else:
        msg += "\n"
    return msg


def _describe(df, field):
    return df[field].describe().iloc[1:]


def _describe_index():
    return ["mean", "std", "min", "25%", "50%", "75%", "max"]



def read(prefix, verbose=True):
    r"""
    Read GEN files into Pandas data frames.

    Parameters
    ----------
    prefix : str
        Path prefix to the set of GEN files.

    Returns
    -------
    sample : dask dataframe
    genotype : dask dataframe

    Examples
    --------
    .. doctest::

        >>> from limix.io.gen import read
        >>> from limix import file_example
        >>>
        >>> with file_example(["example.gen", "example.sample"]) as filepaths:
        ...     data = read(filepaths[0][:-4])
        >>>
        >>> print(data['sample'])
               sample_id subject_id  missing  gender  age  age_of_onset  phenotype_1
        sample
        1A0          1A0       W001  0.00000       2    4            -9            0
        1A1          1A1       W002  0.00000       2    4            -9            0
        1A2          1A2       W003  0.00000       2    4            -9            1
        1A3          1A3       W004  0.09000       2    4            -9            1
        1A4          1A4       W005  0.00000       2    4            -9            1
        >>> print(data['genotype'].head())
                  snp_id  rs_id       pos alleleA alleleB 1A0       1A1       1A2       1A3  \
                                                           AA AB BB  AA AB BB  AA AB BB  AA
        candidate
        SA1          SA1  rs001  10000000       A       G   0  0  1   0  0  1   0  0  1   0
        SA2          SA2  rs002  10010000       A       G   0  0  1   0  1  0   1  0  0   0
        SA3          SA3  rs003  10020000       C       T   1  0  0   0  1  0   0  0  1   0
        SA4          SA4  rs004  10030000       G       T   1  0  0   0  1  0   0  0  1   0
        SA5          SA5  rs005  10040000       C       G   0  0  1   0  1  0   1  0  0   0
        <BLANKLINE>
                                    1A4
                        AB       BB  AA       AB       BB
        candidate
        SA1        0.42770  0.57210   0  0.02070  0.97920
        SA2        1.00000  0.00000   1  0.00000  0.00000
        SA3        0.99670  0.00000   0  0.00000  1.00000
        SA4        1.00000  0.00000   0  0.00000  1.00000
        SA5        1.00000  0.00000   1  0.00000  0.00000
    """

    from pandas import read_csv, MultiIndex

    df_sample = read_csv(prefix + ".sample", header=0, sep=" ", skiprows=[1])

    col_level0_names = ["snp_id", "rs_id", "pos", "alleleA", "alleleB"]
    col_level1_names = [""] * 5
    for s in df_sample.iloc[:, 0]:
        col_level0_names += [s] * 3
        col_level1_names += ["AA", "AB", "BB"]

    tuples = list(zip(col_level0_names, col_level1_names))
    index = MultiIndex.from_tuples(tuples, names=["first", "second"])

    df_sample["sample"] = df_sample["sample_id"]
    df_sample = df_sample.set_index("sample")
    df_sample.index.name = "sample"

    df_gen = read_csv(prefix + ".gen", names=index, sep=" ")
    df_gen["candidate"] = df_gen["snp_id"]
    df_gen = df_gen.set_index("candidate")
    df_gen.index.name = "candidate"

    return dict(sample=df_sample, genotype=df_gen)



from numpy import array, dtype
from numpy.random import RandomState
from numpy.testing import assert_, assert_array_equal, assert_equal
from pandas import DataFrame, Series
from xarray import DataArray

from limix._data import conform_dataset


def test_dataset_conform_dataset():
    y = array([-1.2, 3.4, 0.1])
    samples = ["sample{}".format(i) for i in range(len(y))]

    y = DataFrame(data=y, index=samples)

    random = RandomState(0)

    K = random.randn(3, 4)
    K = K.dot(K.T)
    K = DataFrame(data=K, index=samples, columns=samples)

    M = random.randn(3, 2)
    M = DataFrame(data=M, index=samples)

    G = random.randn(2, 4)
    G = DataFrame(data=G, index=samples[:2])

    data = conform_dataset(y, M=M, K=K)

    assert_array_equal(y.values, data["y"].values)

    y = array([-1.2, 3.4, 0.1, 0.1, 0.0, -0.2])

    data = conform_dataset(DataFrame(data=y, index=samples + samples), M=M, G=G, K=K)

    assert_equal(data["y"].shape, (4, 1))
    assert_equal(data["M"].shape, (4, 2))
    assert_equal(data["G"].shape, (4, 4))
    assert_equal(data["K"].shape, (4, 4))

    samples = ["sample0", "sample1", "sample0", "sample1"]
    assert_array_equal(data["y"].sample, samples)
    assert_array_equal(data["M"].sample, samples)
    assert_array_equal(data["G"].sample, samples)
    assert_array_equal(data["K"].sample_0, samples)
    assert_array_equal(data["K"].sample_1, samples)

    assert_array_equal(data["M"].covariate, [0, 1])
    assert_array_equal(data["G"].candidate, [0, 1, 2, 3])


def test_dataset_pandas_xarray_dask():
    import numpy as np
    import dask.array as da
    import dask.dataframe as dd
    import pandas as pd
    from limix._data import asarray

    x = []

    x.append([1.0, 2.0, 3.0])
    x.append(np.asarray([1.0, 2.0, 3.0]))
    x.append(np.asarray([[1.0], [2.0], [3.0]]))
    x.append(np.asarray([[1], [2], [3]], dtype=int))
    x.append(da.from_array(x[0], 2))
    x.append(da.from_array(x[1], 2))
    x.append(da.from_array(x[2], 2))
    x.append(da.from_array(x[3], 2))

    n = len(x)
    for i in range(n):
        if isinstance(x[i], da.Array):
            tmp = np.asarray(x[i])
            if tmp.ndim == 2:
                tmp = tmp.ravel()
                x.append(dd.from_array(tmp))
            else:
                x.append(dd.from_array(x[i]))
        else:
            tmp = np.asarray(x[i])
            if tmp.ndim == 2:
                tmp = tmp.ravel()
                x.append(pd.Series(tmp))
            else:
                x.append(pd.Series(x[i]))

    for i in range(n):
        if isinstance(x[i], da.Array):
            x.append(dd.from_array(x[i]))
        elif isinstance(x[i], np.ndarray):
            x.append(pd.DataFrame(x[i]))

    n = len(x)

    for i in range(n):
        x.append(DataArray(x[i]))
        x.append(x[-1].chunk(2))

    for xi in x:
        y = asarray(xi, "trait", ["sample", "trait"])
        assert_equal(y.dtype, dtype("float64"))
        assert_array_equal(y.shape, (3, 1))
        assert_(isinstance(y, DataArray))
        if isinstance(xi, Series):
            assert_array_equal(list(xi.index), list(y.coords["sample"].values))
        if isinstance(xi, DataFrame):
            assert_array_equal(list(xi.columns), list(y.coords["trait"].values))

        is_dask = (
            hasattr(xi, "chunks")
            and xi.chunks is not None
            or hasattr(xi, "values")
            and hasattr(xi, "values")
            and hasattr(xi.values, "chunks")
            and xi.values.chunks is not None
        )

        assert_equal(is_dask, y.chunks is not None)
        assert_array_equal(np.asarray(xi).ravel(), np.asarray(y).ravel())


def test_dataset_different_size():
    random = RandomState(0)
    n0 = 5
    n1 = 3
    y = random.randn(n0)
    samples = ["sample{}".format(i) for i in range(len(y))]
    y = DataFrame(data=y, index=samples)

    G = random.randn(n1, 10)

    data = conform_dataset(y, G=G)

    assert_array_equal(data["y"].values, y[:n1])
    assert_array_equal(data["G"].values, G[:n1, :])

    n0 = 3
    n1 = 5
    y = random.randn(n0)
    samples = ["sample{}".format(i) for i in range(len(y))]
    y = DataFrame(data=y, index=samples)

    G = random.randn(n1, 10)

    data = conform_dataset(y, G=G)

    assert_array_equal(data["y"].values, y[:n0])
    assert_array_equal(data["G"].values, G[:n0, :])



"""
Limix package
=============

A flexible and fast generalised mixed model toolbox.

The official documentation together with examples and tutorials can be found
at https://limix.readthedocs.io/.
"""
from . import __config__, her, io, plot, qc, qtl, sh, stats, threads, vardec
from ._cli import cli
from ._config import config
from ._example import file_example
from ._testit import test

__version__ = "3.0.1"


__all__ = [
    "__config__",
    "__version__",
    "cli",
    "config",
    "file_example",
    "gwas",
    "her",
    "io",
    "main",
    "plot",
    "qc",
    "qtl",
    "sh",
    "stats",
    "test",
    "threads",
    "vardec",
    "model",
]



from ._allele import compute_maf
from ._boxcox import boxcox
from ._covariance import normalise_covariance
from ._impute import mean_impute
from ._ld import indep_pairwise
from ._linalg import remove_dependent_cols
from ._mean_std import mean_standardize
from ._missing import count_missingness
from ._quant_gauss import quantile_gaussianize
from ._unique import unique_variants

__all__ = [
    "boxcox",
    "compute_maf",
    "count_missingness",
    "indep_pairwise",
    "mean_impute",
    "mean_standardize",
    "normalise_covariance",
    "quantile_gaussianize",
    "remove_dependent_cols",
    "unique_variants",
]



def asarray(x, target, dims=None):
    from xarray import DataArray
    from limix._bits.dask import is_dataframe as is_dask_dataframe
    from limix._bits.dask import is_array as is_dask_array
    from limix._bits.dask import is_series as is_dask_series
    from limix._bits.dask import array_shape_reveal
    from limix._bits.xarray import is_dataarray
    from ._conf import CONF
    from numpy import issubdtype, integer

    if target not in CONF["targets"]:
        raise ValueError(f"Unknown target name: {target}.")

    import dask.array as da
    import xarray as xr

    if is_dask_dataframe(x) or is_dask_series(x):
        xidx = x.index.compute()
        x = da.asarray(x)
        x = array_shape_reveal(x)
        x0 = xr.DataArray(x)
        x0.coords[x0.dims[0]] = xidx
        if is_dask_dataframe(x):
            x0.coords[x0.dims[1]] = x.columns
        x = x0
    elif is_dask_array(x):
        x = array_shape_reveal(x)
        x = xr.DataArray(x)

    if not is_dataarray(x):
        x = DataArray(x)

    x.name = target

    while x.ndim < 2:
        rdims = set(CONF["data_dims"]["trait"]).intersection(set(x.coords.keys()))
        rdims = rdims - set(x.dims)
        if len(rdims) == 1:
            dim = rdims.pop()
        else:
            dim = "dim_{}".format(x.ndim)
        x = x.expand_dims(dim, x.ndim)

    if isinstance(dims, (tuple, list)):
        dims = {a: n for a, n in enumerate(dims)}
    dims = _numbered_axes(dims)
    if len(set(dims.values())) < len(dims.values()):
        raise ValueError("`dims` must not contain duplicated values.")

    x = x.rename({x.dims[axis]: name for axis, name in dims.items()})
    x = _set_missing_dim(x, CONF["data_dims"][target])
    x = x.transpose(*CONF["data_dims"][target])

    if issubdtype(x.dtype, integer):
        x = x.astype(float)

    for dim in x.dims:
        if x.coords[dim].dtype.kind in {"U", "S"}:
            x.coords[dim].values = x.coords[dim].values.astype(object)

    return x


def _numbered_axes(dims):

    if dims is None:
        return {}

    naxes = {}
    for a, v in dims.items():
        if a == "row":
            naxes[0] = v
        elif a == "col":
            naxes[1] = v
        else:
            naxes[a] = v

    return naxes


def _set_missing_dim(arr, dims):
    unk_dims = set(arr.dims) - set(dims)
    if len(unk_dims) > 1:
        raise ValueError("Too many unknown dimension names.")
    elif len(unk_dims) == 1:
        known_dims = set(dims) - set(arr.dims)
        if len(known_dims) != 1:
            raise ValueError("Can't figure out what is the missing dimension name.")
        arr = arr.rename({unk_dims.pop(): known_dims.pop()})
    return arr



def pca(X, ncomp):
    """
    Principal component analysis.

    Parameters
    ----------
    X : array_like
        Samples-by-dimensions array.
    ncomp : int
        Number of components.

    Returns
    -------
    components : ndarray
        First components ordered by explained variance.
    explained_variance : ndarray
        Explained variance.
    explained_variance_ratio : ndarray
        Percentage of variance explained.

    Examples
    --------
    .. doctest::

        >>> from numpy import round
        >>> from numpy.random import RandomState
        >>> from limix.stats import pca
        >>>
        >>> X = RandomState(1).randn(4, 5)
        >>> r = pca(X, ncomp=2)
        >>> r['components']
        array([[-0.75015369,  0.58346541, -0.07973564,  0.19565682, -0.22846925],
               [ 0.48842769,  0.72267548,  0.01968344, -0.46161623, -0.16031708]])
        >>> r['explained_variance'] # doctest: +FLOAT_CMP
        array([6.44655993, 0.51454938])
        >>> r['explained_variance_ratio'] # doctest: +FLOAT_CMP
        array([0.92049553, 0.07347181])
    """
    from sklearn.decomposition import PCA
    from numpy import asarray

    X = asarray(X, float)
    pca = PCA(n_components=ncomp)
    pca.fit(X)

    return dict(
        components=pca.components_,
        explained_variance=pca.explained_variance_,
        explained_variance_ratio=pca.explained_variance_ratio_,
    )



import scipy.stats as st
from numpy import exp, eye, zeros
from numpy.random import RandomState

from limix.qc import normalise_covariance
from limix.qtl import iscan


def _test_qtl_iscan_three_hypotheses(lik):
    random = RandomState(0)
    n = 10
    ncovariates = 3

    M = random.randn(n, ncovariates)

    E0 = random.randint(0, 2, (n, 1)).astype(float)
    E1 = random.randint(0, 2, (n, 2)).astype(float)

    G = random.randn(n, 4)

    K = random.randn(n, n + 1)
    K = normalise_covariance(K @ K.T)

    beta = random.randn(ncovariates)
    alpha0 = random.randn(E0.shape[1])
    alpha1 = random.randn(E1.shape[1])

    mvn = random.multivariate_normal
    y = _normalize(M @ beta) + _normalize(E0 @ alpha0) + _normalize(E1 @ alpha1)
    y += _normalize(mvn(zeros(n), K + 0.2 * eye(n)))

    idx = [[0, 1], 2, [3]]

    if lik == "poisson":
        y = random.poisson(exp(y))
    elif lik == "bernoulli":
        y = random.binomial(1, 1 / (1 + exp(-y)))
    elif lik == "probit":
        y = random.binomial(1, st.norm.cdf(y))
    elif lik == "binomial":
        ntrials = random.randint(0, 30, len(y))
        y = random.binomial(ntrials, 1 / (1 + exp(-y)))
        lik = (lik, ntrials)

    r = iscan(G, y, lik=lik, idx=idx, K=K, M=M, E0=E0, E1=E1, verbose=False)
    str(r)

    r = iscan(G, y, lik=lik, idx=idx, K=K, M=M, E1=E1, verbose=False)
    str(r)


def test_qtl_iscan_three_hypotheses():
    _test_qtl_iscan_three_hypotheses("normal")
    _test_qtl_iscan_three_hypotheses("poisson")
    _test_qtl_iscan_three_hypotheses("bernoulli")
    _test_qtl_iscan_three_hypotheses("probit")
    _test_qtl_iscan_three_hypotheses("binomial")


def test_qtl_iscan_two_hypotheses():
    random = RandomState(4)
    n = 10
    ncovariates = 3

    M = random.randn(n, ncovariates)

    E1 = random.randint(0, 2, (n, 2)).astype(float)

    G = random.randn(n, 4)

    K = random.randn(n, n + 1)
    K = normalise_covariance(K @ K.T)

    beta = random.randn(ncovariates)
    alpha1 = random.randn(E1.shape[1])

    mvn = random.multivariate_normal
    y = _normalize(M @ beta) + _normalize(E1 @ alpha1)
    y += _normalize(mvn(zeros(n), K + eye(n)))

    idx = [[0, 1], 2, [3]]
    r = iscan(G, y, idx=idx, K=K, M=M, E1=E1, verbose=False)
    str(r)

    r = iscan(G, y, K=K, M=M, E1=E1, verbose=False)
    str(r)


def _normalize(x):
    return (x - x.mean()) / x.std()



def multipletests(pvals, alpha=0.05, method="hs", is_sorted=False):
    """
    Test results and p-value correction for multiple tests.

    Parameters
    ----------
    pvals : array_like
        Uncorrected p-values.
    alpha : float
        FWER, family-wise error rate, e.g. ``0.1``.
    method : string
        Method used for testing and adjustment of pvalues. Can be either the
        full name or initial letters. Available methods are ::

        `bonferroni` : one-step correction
        `sidak` : one-step correction
        `holm-sidak` : step down method using Sidak adjustments
        `holm` : step-down method using Bonferroni adjustments
        `simes-hochberg` : step-up method  (independent)
        `hommel` : closed method based on Simes tests (non-negative)
        `fdr_bh` : Benjamini/Hochberg  (non-negative)
        `fdr_by` : Benjamini/Yekutieli (negative)
        `fdr_tsbh` : two stage fdr correction (non-negative)
        `fdr_tsbky` : two stage fdr correction (non-negative)
    is_sorted : bool
        If ``False`` (default), the p_values will be sorted, but the corrected
        pvalues are in the original order. If ``True``, then it assumed that
        the pvalues are already sorted in ascending order.

    Returns
    -------
    reject : ndarray, boolean
        ``True`` for hypothesis that can be rejected for given alpha.
    pvals_corrected : ndarray
        P-values corrected for multiple tests.
    alphacSidak : float
        Corrected alpha for Sidak method.
    alphacBonf : float
        Corrected alpha for Bonferroni method.

    Notes
    -----
    This is a wrapper around a function from the `statsmodels`_ package.

    .. _statsmodels: http://www.statsmodels.org
    """

    from statsmodels.sandbox.stats.multicomp import multipletests as mt

    return mt(pvals, alpha=alpha, method=method, is_sorted=is_sorted)


def empirical_pvalues(xt, x0):
    """
    Function to compute empirical p-values.

    Compute empirical p-values from the test statistics
    observed on the data and the null test statistics
    (from permutations, parametric bootstraps, etc).

    Parameters
    ----------
    xt : array_like
        Test statistcs observed on data.
    x0 : array_like
        Null test statistcs. The minimum p-value that can be
        estimated is ``1./len(x0)``.

    Returns
    -------
    pvalues : ndarray
        Estimated empirical p-values.

    Examples
    --------
    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.stats import empirical_pvalues
        >>>
        >>> random = RandomState(1)
        >>> x0 = random.chisquare(1, 5)
        >>> x1 = random.chisquare(1, 10000)
        >>>
        >>> empirical_pvalues(x0, x1) # doctest: +FLOAT_CMP
        array([0.563 , 1.    , 0.839 , 0.7982, 0.5803])
    """
    from numpy import argsort, zeros, asarray

    xt = asarray(xt, float)
    x0 = asarray(x0, float)

    idxt = argsort(xt)[::-1]
    idx0 = argsort(x0)[::-1]
    xts = xt[idxt]
    x0s = x0[idx0]
    it = 0
    i0 = 0
    _count = 0
    count = zeros(xt.shape[0])
    while True:
        if x0s[i0] > xts[it]:
            _count += 1
            i0 += 1
            if i0 == x0.shape[0]:
                count[idxt[it:]] = _count
                break
        else:
            count[idxt[it]] = _count
            it += 1
            if it == xt.shape[0]:
                break
    pv = (count + 1) / float(x0.shape[0])
    pv[pv > 1.0] = 1.0
    return pv



from numpy import concatenate, exp, eye, ones, zeros
from numpy.random import RandomState
from numpy.testing import assert_allclose

from limix.stats import multivariate_normal as mvn
from limix.vardec import VarDec


def test_vardec():
    random = RandomState(0)
    nsamples = 20

    X = random.randn(nsamples, 2)
    X = (X - X.mean(0)) / X.std(0)
    X = concatenate((ones((nsamples, 1)), X), axis=1)
    lik = "normal"

    K0 = random.randn(nsamples, 10)
    K0 = K0 @ K0.T
    K0 /= K0.diagonal().mean()
    K0 += eye(nsamples) * 1e-4

    K1 = random.randn(nsamples, 10)
    K1 = K1 @ K1.T
    K1 /= K1.diagonal().mean()
    K1 += eye(nsamples) * 1e-4

    y = X @ random.randn(3) + mvn(random, zeros(nsamples), K0)
    y += mvn(random, zeros(nsamples), K1)

    vardec = VarDec(y, lik, X)
    vardec.append(K0)
    vardec.append(K1)
    vardec.append_iid()

    vardec.fit(verbose=False)
    assert_allclose(vardec.covariance[0].scale, 0.38473522809891697)
    assert_allclose(vardec.covariance[1].scale, 1.1839796169221422)
    assert_allclose(vardec.covariance[2].scale, 2.061153622438558e-09, atol=1e-5)
    assert_allclose(vardec.lml(), -21.91827344966165)
    assert_allclose(
        vardec.effsizes, [-0.5008873352111712, -1.193536406235688, -0.28254298530554534]
    )


def test_vardec_2_matrices():
    random = RandomState(0)
    nsamples = 20

    X = random.randn(nsamples, 2)
    X = (X - X.mean(0)) / X.std(0)
    X = concatenate((ones((nsamples, 1)), X), axis=1)
    lik = "normal"

    K = random.randn(nsamples, 10)
    K = K @ K.T
    K /= K.diagonal().mean()
    K += eye(nsamples) * 1e-4

    y = X @ random.randn(3) + mvn(random, zeros(nsamples), K) + random.randn(nsamples)

    vardec = VarDec(y, lik, X)
    vardec.append(K)
    vardec.append_iid()

    vardec.fit(verbose=False)
    assert_allclose(vardec.covariance[0].scale, 0.32199728815536727, rtol=1e-5)
    assert_allclose(vardec.covariance[1].scale, 1.4182987383374532, rtol=1e-5)
    assert_allclose(vardec.lml(), -33.63946372828994, rtol=1e-5)


def test_vardec_poisson():
    random = RandomState(0)
    nsamples = 20

    X = random.randn(nsamples, 2)
    X = (X - X.mean(0)) / X.std(0)
    X = concatenate((ones((nsamples, 1)), X), axis=1)
    lik = "poisson"

    K0 = random.randn(nsamples, 10)
    K0 = K0 @ K0.T
    K0 /= K0.diagonal().mean()
    K0 += eye(nsamples) * 1e-4

    K1 = random.randn(nsamples, 10)
    K1 = K1 @ K1.T
    K1 /= K1.diagonal().mean()
    K1 += eye(nsamples) * 1e-4

    y = X @ random.randn(3) + mvn(random, zeros(nsamples), K0)
    y += mvn(random, zeros(nsamples), K1)
    y = exp((y - y.mean()) / y.std())

    vardec = VarDec(y, lik, X)
    vardec.append(K0)
    vardec.append(K1)
    vardec.append_iid()

    vardec.fit(verbose=False)
    assert_allclose(vardec.covariance[0].scale, 2.808478303826397e-09, atol=1e-5)
    assert_allclose(vardec.covariance[1].scale, 0.3503800007985209)
    assert_allclose(vardec.covariance[2].scale, 2.061153622438558e-09, atol=1e-5)
    assert_allclose(vardec.lml(), -28.887285796984564)


def test_vardec_poisson_2_matrices():
    random = RandomState(0)
    nsamples = 20

    X = random.randn(nsamples, 2)
    X = (X - X.mean(0)) / X.std(0)
    X = concatenate((ones((nsamples, 1)), X), axis=1)
    lik = "poisson"

    K = random.randn(nsamples, 10)
    K = K @ K.T
    K /= K.diagonal().mean()
    K += eye(nsamples) * 1e-4

    y = X @ random.randn(3) + mvn(random, zeros(nsamples), K)
    y = exp((y - y.mean()) / y.std())

    vardec = VarDec(y, lik, X)
    vardec.append(K)
    vardec.append_iid()

    vardec.fit(verbose=False)
    assert_allclose(vardec.covariance[0].scale, 0.10726852397002325, atol=1e-5)
    assert_allclose(vardec.covariance[1].scale, 4.168569936272955e-11, atol=1e-5)
    assert_allclose(vardec.lml(), -26.36419072811823)



import sys

from limix._display import session_line

from .._bits import unvec
from .._data import asarray as _asarray, conform_dataset, normalize_likelihood
from .._display import session_block
from ._assert import assert_finite
from ._result import IScanResultFactory


def iscan(G, y, lik="normal", K=None, M=None, idx=None, E0=None, E1=None, verbose=True):
    r"""
    Single-trait association with interaction test via generalized linear mixed models.

    The general formulae for normally distributed traits is

    .. math::

        𝐲 = 𝙼𝛂 + (𝙶⊙𝙴₀)𝛃₀ + (𝙶⊙𝙴₁)𝛃₁ + 𝐮 + 𝛆,\\
        \text{where}~~ 𝐮∼𝓝(𝟎, 𝓋₀𝙺) ~~\text{and}~~ 𝛆∼𝓝(𝟎, 𝓋₁𝙸).

    The operator ⊙ works as follows:

    .. math::

        𝙰⊙𝙱 = [𝙰₀𝙱₀ ~~...~~ 𝙰₀𝙱ₙ ~~ 𝙰₁𝙱₀ ~~...~~ 𝙰₁𝙱ₙ ~~...~~ 𝙰ₘ𝙱ₙ]

    The covariates is enconded in matrix 𝙼 while the candidate set is enconded in matrix
    𝙶. The parameters are the effect sizes 𝛂, 𝛃₀, and 𝛃₁, and the variances 𝓋₀ and 𝓋₁.

    It performs likelihood-ratio tests for the following cases, where the first
    hypothesis is the null one while the second hypothesis is the alternative one:

    - H₀ vs H₁: testing for vec(𝛃₀) ≠ 𝟎 while vec(𝛃₁) = 𝟎
    - H₀ vs H₂: testing for [vec(𝛃₀) vec(𝛃₁)] ≠ 𝟎
    - H₁ vs H₂: testing for vec(𝛃₁) ≠ 𝟎

    It also supports generalized linear mixed models (GLMM). In this case, the following
    likelihoods are implemented:
    - Bernoulli
    - Probit
    - Binomial
    - Poisson

    Formally, let p(𝜇) be one of the supported probability distributions where 𝜇 is
    its mean. The H₀ model is defined as follows:

    .. math::

        yᵢ ∼ p(𝜇ᵢ=g(zᵢ)) ~~\text{for}~~ 𝐳 ∼ 𝓝(𝙼𝛂 + (𝙶⊙𝙴₀)𝛃₀ + (𝙶⊙𝙴₁)𝛃₁, 𝓋₀𝙺 + 𝓋₁𝙸).

    g(⋅) is the corresponding canonical link function for the Bernoulli, Binomial, and
    Poisson likelihoods. The Probit likelihood, on the other hand, is a Bernoulli
    likelihood with probit link function.

    Parameters
    ----------
    G : n×m array_like
        Genetic candidates.
    Y : n×p array_like
        Rows are samples and columns are phenotypes.
    lik : tuple, "normal", "bernoulli", "probit", "binomial", "poisson"
        Sample likelihood describing the residual distribution.
        Either a tuple or a string specifying the likelihood is required. The Normal,
        Bernoulli, Probit, and Poisson likelihoods can be selected by providing a
        string. Binomial likelihood on the other hand requires a tuple because of the
        number of trials: ``("binomial", array_like)``. Defaults to ``"normal"``.
    K : n×n array_like
        Sample covariance, often the so-called kinship matrix.
    M : n×c array_like
        Covariates matrix.
    idx : list
        List of candidate indices that defines the set of candidates to be used in the
        tests.
    E0 : array_like
        Matrix representing the first environment.
    E1 : array_like
        Matrix representing the second environment.
    verbose : bool, optional
        ``True`` to display progress and summary; ``False`` otherwise.

    Returns
    -------
    result : :class:`limix.qtl._result.IScanResult`
        P-values, log of marginal likelihoods, effect sizes, and associated statistics.

    Notes
    -----
    It will raise a ``ValueError`` exception if non-finite values are passed. Please,
    refer to the :func:`limix.qc.mean_impute` function for missing value imputation.
    """
    from numpy_sugar.linalg import economic_qs
    from xarray import concat
    from numpy import asarray, empty, ones

    lik = normalize_likelihood(lik)
    lik_name = lik[0]

    with session_block("QTL analysis", disable=not verbose):

        with session_line("Normalising input... ", disable=not verbose):

            data = conform_dataset(y, M, G=G, K=K)

        Y = data["y"]
        M = data["M"]
        G = data["G"]
        K = data["K"]

        assert_finite(y, M, K)
        nsamples = y.shape[0]

        if E1 is None:
            E1 = ones((nsamples, 1))

        if E0 is None:
            E0 = empty((nsamples, 0))

        E0 = _asarray(E0, "env0", ["sample", "env"])
        E1 = _asarray(E1, "env1", ["sample", "env"])
        E01 = concat([E0, E1], dim="env")

        if K is not None:
            QS = economic_qs(K)
        else:
            QS = None

        if lik_name == "normal":
            scanner, v0, v1 = _lmm(Y.values.ravel(), M.values, QS, verbose)
        else:
            scanner, v0, v1 = _glmm(Y.values.ravel(), lik, M.values, QS, verbose)

        r = IScanResultFactory(
            lik_name,
            Y.trait,
            M.covariate,
            G.candidate,
            E0.env,
            E1.env,
            scanner.null_lml,
            scanner.null_beta,
            scanner.null_beta_se,
            v0,
            v1,
        )

        if idx is None:

            assert E1.shape[1] > 0
            idx = range(G.shape[1])

            if E0.shape[1] == 0:
                r1 = scanner.fast_scan(G, verbose)

            for i in idx:
                i = _2d_sel(i)
                g = asarray(G[:, i], float)

                if E0.shape[1] > 0:
                    r1 = scanner.scan(g, E0)
                    h1 = _normalise_scan_names(r1)
                else:
                    h1 = _normalise_scan_names({k: v[i] for k, v in r1.items()})
                    h1["covariate_effsizes"] = h1["covariate_effsizes"].ravel()
                    h1["covariate_effsizes_se"] = h1["covariate_effsizes_se"].ravel()

                r2 = scanner.scan(g, E01)
                h2 = _normalise_scan_names(r2)
                r.add_test(i, h1, h2)
        else:
            for i in idx:
                i = _2d_sel(i)
                g = asarray(G[:, i], float)

                r1 = scanner.scan(g, E0)
                r2 = scanner.scan(g, E01)

                h1 = _normalise_scan_names(r1)
                h2 = _normalise_scan_names(r2)
                r.add_test(i, h1, h2)

        r = r.create()
        if verbose:
            print(r)

        return r


def _normalise_scan_names(r):
    return {
        "lml": r["lml"],
        "covariate_effsizes": r["effsizes0"],
        "covariate_effsizes_se": r["effsizes0_se"],
        "candidate_effsizes": r["effsizes1"],
        "candidate_effsizes_se": r["effsizes1_se"],
        "scale": r["scale"],
    }


def _2d_sel(idx):
    from collections.abc import Iterable

    if not isinstance(idx, (slice, Iterable)):
        return [idx]

    return idx


class ScannerWrapper:
    def __init__(self, scanner):
        self._scanner = scanner

    @property
    def null_lml(self):
        return self._scanner.null_lml()

    @property
    def null_beta(self):
        return self._scanner.null_beta

    @property
    def null_beta_se(self):
        from numpy import sqrt

        se = sqrt(self._scanner.null_beta_covariance.diagonal())
        return se

    def fast_scan(self, G, verbose):
        r = self._scanner.fast_scan(G, verbose=verbose)
        r["effsizes1"] = unvec(r["effsizes1"], (-1, G.shape[1])).T
        r["effsizes1_se"] = unvec(r["effsizes1_se"], (-1, G.shape[1])).T
        return r

    def scan(self, G, E):
        from .._bits import cdot

        r = self._scanner.scan(cdot(G, E))
        r["effsizes1"] = unvec(r["effsizes1"], (-1, G.shape[1])).T
        r["effsizes1_se"] = unvec(r["effsizes1_se"], (-1, G.shape[1])).T
        return r


def _lmm(y, M, QS, verbose):
    from glimix_core.lmm import LMM

    lmm = LMM(y, M, QS, restricted=False)
    lmm.fit(verbose=verbose)
    sys.stdout.flush()

    if QS is None:
        v0 = None
    else:
        v0 = lmm.v0
    v1 = lmm.v1
    scanner = ScannerWrapper(lmm.get_fast_scanner())

    return scanner, v0, v1


def _glmm(y, lik, M, QS, verbose):
    from glimix_core.glmm import GLMMExpFam, GLMMNormal

    glmm = GLMMExpFam(y.ravel(), lik, M, QS)

    glmm.fit(verbose=verbose)
    v0 = glmm.v0
    v1 = glmm.v1
    sys.stdout.flush()

    eta = glmm.site.eta
    tau = glmm.site.tau

    gnormal = GLMMNormal(eta, tau, M, QS)
    gnormal.fit(verbose=verbose)

    scanner = ScannerWrapper(gnormal.get_fast_scanner())

    return scanner, v0, v1



import warnings

from limix._bits import dask, get_shape, numpy, pandas, xarray


def mean_impute(X, axis=-1, inplace=False):
    """
    Impute ``NaN`` values.

    It defaults to column-wise imputation.

    Parameters
    ----------
    X : array_like
        Matrix to have its missing values imputed.
    axis : int, optional
        Axis value. Defaults to `-1`.
    inplace : bool, optional
        Defaults to `False`.

    Returns
    -------
    ndarray
        Imputed array.

    Examples
    --------
    .. doctest::

        >>> from numpy.random import RandomState
        >>> from numpy import nan, array_str
        >>> from limix.qc import mean_impute
        >>>
        >>> random = RandomState(0)
        >>> X = random.randn(5, 2)
        >>> X[0, 0] = nan
        >>>
        >>> print(array_str(mean_impute(X), precision=4))
        [[ 0.9233  0.4002]
         [ 0.9787  2.2409]
         [ 1.8676 -0.9773]
         [ 0.9501 -0.1514]
         [-0.1032  0.4106]]

    .. _Dask: https://dask.pydata.org/
    """
    from numpy import asarray

    if isinstance(X, (tuple, list)):
        if inplace:
            raise ValueError("Can't use `inplace=True` for {}.".format(type(X)))
        X = asarray(X, float)

    if numpy.is_array(X):
        X = _impute_numpy(X, axis, inplace)
    elif pandas.is_series(X):
        X = _impute_pandas_series(X, axis, inplace)
    elif pandas.is_dataframe(X):
        X = _impute_pandas_dataframe(X, axis, inplace)
    elif dask.is_array(X):
        X = _impute_dask_array(X, axis, inplace)
    elif dask.is_series(X):
        raise NotImplementedError()
    elif dask.is_dataframe(X):
        X = _impute_dask_dataframe(X, axis, inplace)
    elif xarray.is_dataarray(X):
        X = _impute_xarray_dataarray(X, axis, inplace)
    else:
        raise NotImplementedError()

    return X


def _impute_numpy(X, axis, inplace):
    from numpy import isnan, nanmean

    orig_shape = get_shape(X)
    if X.ndim == 1:
        X = X.reshape(orig_shape + (1,))

    if not inplace:
        X = X.copy()

    X = X.swapaxes(-1, axis)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        m = nanmean(X, axis=0)
    for i, mi in enumerate(m):
        X[isnan(X[:, i]), i] = mi

    X = X.swapaxes(-1, axis)

    return X.reshape(orig_shape)


def _impute_pandas_series(x, axis, inplace):
    if not inplace:
        x = x.copy()

    a = x.to_numpy()
    _impute_numpy(a, axis, True)
    x[:] = a

    return x


def _impute_dask_dataframe(x, axis, inplace):
    import dask.dataframe as dd

    if inplace:
        raise NotImplementedError()

    d = x.to_dask_array(lengths=True)
    d = _impute_dask_array(d, axis, False)
    x = dd.from_dask_array(d, columns=x.columns, index=x.index)
    return x


def _impute_pandas_dataframe(x, axis, inplace):
    if not inplace:
        x = x.copy()

    a = x.to_numpy()
    _impute_numpy(a, axis, True)
    x[:] = a

    return x


def _impute_dask_array(x, axis, inplace):
    import dask.array as da

    x = x.swapaxes(-1, axis)
    m = da.nanmean(x, axis=0).compute()
    start = 0

    arrs = []
    for i in range(len(x.chunks[1])):
        end = start + x.chunks[1][i]
        impute = _get_imputer(m[start:end], inplace)
        arrs.append(x[:, start:end].map_blocks(impute, dtype=float))
        start = end
    return da.concatenate(arrs, axis=1).swapaxes(-1, axis)


def _impute_xarray_dataarray(X, axis, inplace):
    if not inplace:
        X = X.copy(deep=True)

    data = X.data

    if dask.is_array(data):
        data = _impute_dask_array(data, axis, inplace)
    else:
        data = _impute_numpy(data, axis, inplace)

    X.data = data
    return X


def _get_imputer(m, inplace):
    from numpy import isnan

    def impute(X):
        A = X.copy()

        isn = isnan(A)
        A[:] = 0
        A[isn] = 1

        if not inplace:
            X = X.copy()

        X[isn] = 0
        X += A * m

        return X

    return impute



"""
Statistics.
"""

from ._allele import allele_expectation, allele_frequency, compute_dosage
from ._chi2 import Chi2Mixture
from ._confusion import confusion_matrix
from ._kinship import linear_kinship
from ._lrt import lrt_pvalues
from ._pca import pca
from ._pvalue import empirical_pvalues, multipletests
from ._random import multivariate_normal

__all__ = [
    "Chi2Mixture",
    "allele_expectation",
    "allele_frequency",
    "compute_dosage",
    "confusion_matrix",
    "empirical_pvalues",
    "linear_kinship",
    "lrt_pvalues",
    "multipletests",
    "multivariate_normal",
    "pca",
]



def pytest_configure(*_):
    _compatibility()
    import doctest
    import matplotlib as mpl

    mpl.use("agg")

    pandas_format()
    doctest.ELLIPSIS_MARKER = "-ignore-"


def pandas_format():
    import pandas as pd

    pd.set_option("display.width", 88)
    pd.set_option("display.max_columns", 79)
    pd.set_option("display.max_rows", 60)
    pd.set_option("display.large_repr", "truncate")
    pd.set_option("display.float_format", "{:8.5f}".format)


def _compatibility():
    import warnings

    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    warnings.filterwarnings("ignore", message="numpy.ufunc size changed")



import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_array_equal as assert_eq, assert_string_equal

from limix._data import asarray


def test_asarray():
    from limix._data._conf import CONF

    random = np.random.RandomState(0)
    arr = random.randn(2, 3)

    def _assert_target(x, target):
        assert_eq(x.dims, CONF["data_dims"][target])
        assert_string_equal(x.name, target)

    _assert_target(asarray(arr, "trait", {"row": "sample"}), "trait")
    _assert_target(asarray(arr, "trait", {"col": "sample"}), "trait")
    _assert_target(asarray(arr, "trait", {"row": "trait"}), "trait")
    _assert_target(asarray(arr, "trait", {"col": "trait"}), "trait")
    _assert_target(asarray(arr, "trait", {"row": "trait", "col": "sample"}), "trait")
    _assert_target(asarray(arr, "trait", {"row": "trait", 1: "sample"}), "trait")
    _assert_target(asarray(arr, "trait", {0: "trait", 1: "sample"}), "trait")
    _assert_target(asarray(arr, "trait", ["trait"]), "trait")
    _assert_target(asarray(arr, "trait", ["sample"]), "trait")
    _assert_target(asarray(arr, "trait", ["sample", "trait"]), "trait")
    _assert_target(asarray(arr, "trait", ["sample", "wrong"]), "trait")

    with pytest.raises(ValueError):
        _assert_target(asarray(arr, "wrong"), "trait")

    with pytest.raises(ValueError):
        _assert_target(asarray(arr, "trait", ["wrong1", "wrong2"]), "trait")

    _assert_target(
        asarray(xr.DataArray(arr, dims=["sample", "trait"]), "trait"), "trait"
    )
    _assert_target(
        asarray(xr.DataArray(arr, dims=["sample", "dim2"]), "trait"), "trait"
    )
    _assert_target(asarray(xr.DataArray(arr, dims=["dim1", "trait"]), "trait"), "trait")
    _assert_target(
        asarray(
            xr.DataArray(arr, dims=["dim1", "trait"]), "trait", ["sample", "trait"]
        ),
        "trait",
    )
    _assert_target(
        asarray(xr.DataArray(arr, dims=["dim1", "trait"]), "trait", {0: "sample"}),
        "trait",
    )
    _assert_target(
        asarray(xr.DataArray(arr, dims=["dim1", "trait"]), "trait", {1: "sample"}),
        "trait",
    )

    with pytest.raises(ValueError):
        _assert_target(
            asarray(xr.DataArray(arr), "trait", {0: "sample", 1: "sample"}), "trait"
        )

    with pytest.raises(ValueError):
        _assert_target(
            asarray(xr.DataArray(arr), "trait", ["sample", "sample"]), "trait"
        )



def read(filename, sep=None, header=True, verbose=True):
    """
    Read a CSV file.

    Parameters
    ----------
    filename : str
        Path to a CSV file.
    sep : str
        Separator. ``None`` triggers auto-detection. Defaults to ``None``.
    header : bool
        ``True`` for file with a header; ``False`` otherwise. Defaults
        to ``True``.
    verbose : bool
        `True` for progress information; `False` otherwise.

    Returns
    -------
    data : dask dataframes

    Examples
    --------
    .. doctest::

        >>> from limix.io.csv import read
        >>> from limix import file_example
        >>>
        >>> with file_example("data.csv") as filepath:
        ...     df = read(filepath, verbose=False)
        ...     print(df)  # doctest: +FLOAT_CMP
           pheno   attr1 attr2 attr3
        0    sex  string    10     a
        1   size   float    -3     b
        2  force     int     f     c
    """
    from dask.dataframe import read_csv as dask_read_csv
    from pandas import read_csv as pandas_read_csv
    from .._display import session_line

    if sep is None:
        sep = _infer_separator(filename)

    header = 0 if header else None

    with session_line("Reading {}... ".format(filename), disable=not verbose):

        if _is_large_file(filename):
            df = dask_read_csv(filename, sep=sep, header=header)
        else:
            df = pandas_read_csv(filename, sep=sep, header=header)

    if len(df.columns) > 0:
        if df.columns[0] == "Unnamed: 0":
            df = df.set_index("Unnamed: 0")
            df.index.name = None

    return df


def _see(filepath, header, verbose=True):
    """
    Shows a human-friendly representation of a CSV file.

    Parameters
    ----------
    filepath : str
        CSV file path.
    header : bool
        ``True`` for parsing the header; ``False`` otherwise.
    verbose : bool
        ``True`` for verbose; ``False`` otherwise.

    Returns
    -------
    str
        CSV representation.
    """
    from pandas import read_csv
    from .._display import session_line

    if header:
        header = 0
    else:
        header = None

    with session_line(desc="Reading %s... " % filepath, disable=not verbose):
        sep = _infer_separator(filepath)
        msg = read_csv(filepath, sep=sep, header=header).head()

    print(msg)


def _count(candidates, line):
    counter = {c: 0 for c in candidates}
    for i in line:
        if i in candidates:
            counter[i] += 1
    return counter


def _update(counter, c):
    for (k, v) in c.items():
        if counter[k] != v:
            del counter[k]


def _infer_separator(fn):
    nmax = 9

    with open(fn, "r") as f:
        line = _remove_repeat(f.readline())
        counter = _count(set(line), line)

        for _ in range(nmax - 1):
            line = _remove_repeat(f.readline())
            if len(line) == 0:
                break
            c = _count(set(counter.keys()), line)
            _update(counter, c)
            if len(counter) == 1:
                return next(iter(counter.keys()))

    for c in set([",", "\t", " "]):
        if c in counter:
            return c

    counter = list(counter.items())
    if len(counter) == 0:
        return None

    counter = sorted(counter, key=lambda kv: kv[1])
    return counter[-1][0]


def _remove_repeat(s):
    from re import sub

    return sub(r"(.)\1+", r"\1", s)


def _is_large_file(filepath):
    import os

    large = 1024 * 1024 * 100
    return os.path.getsize(filepath) >= large



__all__ = ["array_shape_reveal", "is_array", "is_series"]


def array_shape_reveal(a):
    r""" Compute shape from lazy arrays.

    Dask arrays might have unknown dimensions due to performance
    reasons. This function compute those dimensions.
    """
    from numpy import any, isnan
    import dask.array as da

    shape = da.compute(*a.shape)
    if any(isnan(shape)):
        # Rebuild Dask Array with known chunks
        return da.Array(a.__dask_graph__(), a.name, _get_chunks(a), a.dtype)
    return a


def is_array(a):
    pkg = a.__class__.__module__.split(".")[0]
    name = a.__class__.__name__

    return pkg == "dask" and name == "Array"


def is_series(a):
    pkg = a.__class__.__module__.split(".")[0]
    name = a.__class__.__name__

    return pkg == "dask" and name == "Series"


def is_dataframe(a):
    pkg = a.__class__.__module__.split(".")[0]
    name = a.__class__.__name__

    return pkg == "dask" and name == "DataFrame"


def _get_shape_helper(a):
    from numpy import asarray

    s = asarray(a.shape, dtype=int)
    return s[len(s) * (None,) + (slice(None),)]


def _get_all_chunk_shapes(a):
    return a.map_blocks(
        _get_shape_helper,
        dtype=int,
        chunks=tuple(len(c) * (1,) for c in a.chunks) + ((a.ndim,),),
        new_axis=a.ndim,
    )


def _get_chunks(a):
    cs = _get_all_chunk_shapes(a)

    c = []
    for i in range(a.ndim):
        s = a.ndim * [0] + [i]
        s[i] = slice(None)
        s = tuple(s)

        c.append(tuple(cs[s]))

    return tuple(c)



import click

import limix


@click.command()
@click.pass_context
@click.argument("filepath", type=click.Path(exists=True))
def remove(ctx, filepath):
    """Remove a file."""

    limix.sh.remove(filepath)



_max_nthreads = None


def set_max_nthreads(nthreads):
    """ Set the maximum number of threads.

    Parameters
    ----------
    nthreads : int
        Maximum number of threads.
    """
    import dask
    from multiprocessing.pool import ThreadPool

    global _max_nthreads

    nthreads = int(nthreads)
    if nthreads < 1:
        raise ValueError("Cannot set number of threads smaller than one.")
    _max_nthreads = nthreads
    dask.config.set(pool=ThreadPool(_max_nthreads))


def get_max_nthreads():
    """ Get the maximum number of threads

    Returns
    -------
    int
        Maximum number of threads.
    """
    from multiprocessing import cpu_count

    global _max_nthreads

    if _max_nthreads is None:
        return cpu_count()
    return _max_nthreads



__all__ = ["set_coord", "take", "query"]


def set_coord(x, dim, values):
    r""" Assign a new coordinate or subset an existing one. """
    if dim not in x.coords:
        return x.assign_coords(**{dim: list(values)})
    return x.loc[{dim: x.get_index(dim).isin(values)}]


def take(x, indices, dim):
    r""" Subset a data array on an arbitrary dimension. """
    sl = [slice(None)] * x.ndim
    axis = next(i for i, d in enumerate(x.dims) if d == dim)
    sl[axis] = indices
    return x[tuple(sl)]


def query(x, expr):
    from io import StringIO
    from tokenize import generate_tokens, OP, NAME

    expr = expr.strip()
    if len(expr) == 0:
        return x
    tokens = list(generate_tokens(StringIO(expr).readline))

    final_expr = ""
    last = None
    for t in tokens:
        if t.type == NAME and t.string not in ["isin", "None"]:

            is_boolean = last is not None
            is_boolean = is_boolean and not (last.type == OP and _is_comp(last.string))
            is_boolean = is_boolean and _is_boolean(t.string)
            if is_boolean:
                final_expr += _cast_boolean(t.string)
            else:
                final_expr += 'x["{}"]'.format(t.string)
        elif t.type == OP and _is_comp(t.string):
            final_expr += " {} ".format(t.string)
        else:
            final_expr += t.string
        last = t

    return eval("x.where(" + final_expr + ", drop=True)")


def is_dataarray(a):
    pkg = a.__class__.__module__.split(".")[0]
    name = a.__class__.__name__

    return pkg == "xarray" and name == "DataArray"


def _is_comp(v):
    return v in set(["<", ">", "<=", ">=", "==", "!="])


def _is_boolean(v):
    return v.lower() in set(["and", "or", "not"])


def _cast_boolean(v):
    d = {"and": " & ", "or": " | ", "not": "~"}
    return d[v.lower()]



from .._data import conform_dataset, normalize_likelihood
from .._display import session_block, session_line
from ..qtl._assert import assert_finite


def estimate(y, lik, K, M=None, verbose=True):
    """
    Estimate the so-called narrow-sense heritability.

    It supports Normal, Bernoulli, Probit, Binomial, and Poisson phenotypes.

    Parameters
    ----------
    y : array_like
        Array of trait values of n individuals.
    lik : tuple, "normal", "bernoulli", "probit", "binomial", "poisson"
        Sample likelihood describing the residual distribution.
        Either a tuple or a string specifying the likelihood is required. The Normal,
        Bernoulli, Probit, and Poisson likelihoods can be selected by providing a
        string. Binomial likelihood on the other hand requires a tuple because of the
        number of trials: ``("binomial", array_like)``. Defaults to ``"normal"``.
    K : n×n array_like
        Sample covariance, often the so-called kinship matrix. It might be, for example,
        the estimated kinship relationship between the individuals. The provided matrix
        will be normalised as ``K = K / K.diagonal().mean()``.
    M : n×c array_like, optional
        Covariates matrix. If an array is passed, it will used as is; no normalisation
        will be performed. If ``None`` is passed, an offset will be used as the only
        covariate. Defaults to ``None``.
    verbose : bool, optional
        ``True`` to display progress and summary; ``False`` otherwise.

    Returns
    -------
    float
        Estimated heritability.

    Examples
    --------
    .. doctest::

        >>> from numpy import dot, exp, sqrt
        >>> from numpy.random import RandomState
        >>> from limix.her import estimate
        >>>
        >>> random = RandomState(0)
        >>>
        >>> G = random.randn(150, 200) / sqrt(200)
        >>> K = dot(G, G.T)
        >>> z = dot(G, random.randn(200)) + random.randn(150)
        >>> y = random.poisson(exp(z))
        >>>
        >>> print(estimate(y, 'poisson', K, verbose=False))  # doctest: +FLOAT_CMP
        0.18311439918863426

    Notes
    -----
    It will raise a ``ValueError`` exception if non-finite values are passed. Please,
    refer to the :func:`limix.qc.mean_impute` function for missing value imputation.
    """
    from numpy_sugar.linalg import economic_qs
    from numpy import pi, var, diag
    from glimix_core.glmm import GLMMExpFam
    from glimix_core.lmm import LMM

    lik = normalize_likelihood(lik)
    lik_name = lik[0]

    with session_block("Heritability analysis", disable=not verbose):

        with session_line("Normalising input...", disable=not verbose):
            data = conform_dataset(y, M=M, K=K)

        y = data["y"]
        M = data["M"]
        K = data["K"]

        assert_finite(y, M, K)

        if K is not None:
            K = K / diag(K).mean()
            QS = economic_qs(K)
        else:
            QS = None

        if lik_name == "normal":
            method = LMM(y.values, M.values, QS, restricted=True)
            method.fit(verbose=verbose)
        else:
            method = GLMMExpFam(y, lik, M.values, QS, n_int=500)
            method.fit(verbose=verbose, factr=1e6, pgtol=1e-3)

        g = method.scale * (1 - method.delta)
        e = method.scale * method.delta
        if lik_name == "bernoulli":
            e += pi * pi / 3

        v = var(method.mean())

        return g / (v + g + e)



import click

from ._click import OrderedCommand


@click.command(cls=OrderedCommand)
@click.pass_context
@click.argument("trait")
@click.argument("genotype")
@click.option(
    "--covariate",
    help="Specify the file path to a file containing the covariates.",
    default=None,
)
@click.option(
    "--kinship",
    help="Specify the file path to a file containing the kinship matrix.",
    default=None,
)
@click.option(
    "--lik",
    help=(
        "Specify the type of likelihood that will described the residual error "
        "distribution. It can be 'normal', 'bernoulli', 'binomial', or 'poisson'."
    ),
    default="normal",
)
@click.option(
    "--drop-maf",
    help=(
        "Drop out candidates having a minor allele frequency below the provided "
        "threshold."
    ),
)
@click.option(
    "--drop-missing",
    help=("Drop out samples, candidates, or covariates with missing values."),
    multiple=True,
)
@click.option(
    "--where",
    help=(
        "Filtering expression to select which phenotype, genotype loci, and covariates"
        " to use in the analysis. The syntax is <TARGET>:<COND>,"
        " where <COND> is the first argument of the method `DataArray.where`."
    ),
    multiple=True,
)
@click.option(
    "--impute",
    help=(
        "Impute missing values for phenotype, genotype, and covariate. The syntax is"
        " <TARGET>:<DIM>:<METHOD>, where <METHOD> can be 'mean'. Defaults to 'mean'"
    ),
    multiple=True,
)
@click.option(
    "--normalize",
    help=(
        "Normalize phenotype, genotype, and covariate. The syntax is "
        "<TARGET>:<DIM>:<METHOD>, where <METHOD> can be 'gaussianize'"
        " or 'mean_std'. Defaults to 'gaussianize'."
    ),
    multiple=True,
)
@click.option(
    "--output-dir", help="Specify the output directory path.", default="output"
)
@click.option(
    "--verbose/--quiet", "-v/-q", help="Enable or disable verbose mode.", default=True
)
@click.option(
    "--dry-run/--no-dry-run",
    help="Perform a trial run with no scan taking place.",
    default=False,
)
def scan(
    ctx, trait, genotype, covariate, kinship, lik, output_dir, verbose, dry_run, **_
):
    """ Single-variant association testing via mixed models.

    This analysis requires minimally the specification of one phenotype
    (PHENOTYPES_FILE) and genotype data (GENOTYPE_FILE).

    The --filter option allows for selecting a subset of the original dataset for
    the analysis. For example,

        --filter="genotype: (chrom == '3') & (pos > 100) & (pos < 200)"

    states that only loci of chromosome 3 having a position inside the range (100, 200)
    will be considered. The --filter option can be used multiple times in the same
    call. In general, --filter accepts a string of the form

        <DATA-TYPE>: <BOOL-EXPR>

    where <DATA-TYPE> can be phenotype, genotype, or covariate. <BOOL-EXPR> is a boolean
    expression involving row or column names. Please, consult `pandas.DataFrame.query`
    function from Pandas package for further information.
    \f

    Examples
    --------

    ... doctest::

        # First we perform a quick file inspection. This step is optional but is very
        # useful to check whether `limix` is able to read them and print out their
        # metadata.
        limix show phenotypes.csv
        limix show genotype.bgen
        limix show kinship.raw

        # We now perform the analysis, specifying the genotype loci and the phenotype
        # of interest.
        limix phenotypes.csv genotype.bgen --kinship-file=kinship.raw \
            --output-dir=results \
            --filter="phenotype: col == 'height'" \
            --filter="genotype: (chrom == '3') & (pos > 100) & (pos < 200)"
    """
    import sys
    from os import makedirs
    from os.path import abspath, exists, join
    import traceback
    from limix._display import session_block, banner, session_line, indent, print_exc
    from limix.qtl import scan
    from limix.io import fetch
    from .pipeline import Pipeline
    from limix._data import conform_dataset
    from .preprocess import impute as impute_func
    from .preprocess import normalize as normalize_func
    from .preprocess import where as where_func
    from .preprocess import drop_missing, drop_maf

    print(banner())

    if ctx.obj is None:
        ctx.obj = {"preprocess": []}

    output_dir = abspath(output_dir)
    if not dry_run:
        if not exists(output_dir):
            makedirs(output_dir, exist_ok=True)

    def _print_data_array(arr, verbose):
        if verbose:
            print("\n{}\n".format(indent(_clean_data_array_repr(arr))))

    data = {"y": None, "G": None, "K": None}

    data["y"] = fetch("trait", trait, verbose)
    _print_data_array(data["y"], verbose)

    data["G"] = fetch("genotype", genotype, verbose)
    _print_data_array(data["G"], verbose)

    if covariate is not None:
        data["M"] = fetch("covariate", covariate, verbose)
        _print_data_array(data["M"], verbose)

    if kinship is not None:
        data["K"] = fetch("kinship", kinship, verbose)
        _print_data_array(data["K"], verbose)

    with session_line("Matching samples... "):
        data = conform_dataset(**data)
    data = {k: v for k, v in data.items() if v is not None}

    if data["y"].sample.size == 0:
        raise RuntimeError(
            "Exiting early because there is no sample left after matching samples."
            + " Please, check your sample ids."
        )

    oparams = _ordered_params(ctx)

    with session_block("preprocessing", disable=not verbose):
        pipeline = Pipeline(data)
        preproc_params = [
            i
            for i in oparams
            if i[0] in ["impute", "normalize", "where", "drop_missing", "drop_maf"]
        ]

        for p in preproc_params:
            if p[0] == "where":
                pipeline.append(where_func, "where", p[1])
            elif p[0] == "normalize":
                pipeline.append(normalize_func, "normalize", p[1])
            elif p[0] == "impute":
                pipeline.append(impute_func, "impute", p[1])
            elif p[0] == "drop_maf":
                pipeline.append(drop_maf, "drop-maf", p[1])
            elif p[0] == "drop_missing":
                pipeline.append(drop_missing, "drop-missing", p[1])

        data = pipeline.run()

    if dry_run:
        print("Exiting early because of dry-run option.")
        return

    if "K" not in data:
        data["K"] = None
    try:
        res = scan(
            data["G"], data["y"], lik=lik, K=data["K"], M=data["M"], verbose=verbose
        )
    except Exception as e:
        print_exc(traceback.format_stack(), e)
        sys.exit(1)

    with session_line("Saving results to `{}`... ".format(output_dir)):
        res.to_csv(join(output_dir, "null.csv"), join(output_dir, "alt.csv"))


def _clean_data_array_repr(arr):
    txt = str(arr)
    txt = txt.replace("xarray.DataArray ", "")
    txt = txt.replace("object ", "")
    txt = txt.replace("int64 ", "")
    txt = txt.replace("<U5 ", "")
    txt = txt.replace("dask.array", "array")
    return txt


def _ordered_params(ctx):
    args_seq = []
    for p in ctx.param_order:
        for opt, val in ctx.params.items():
            if p.name == opt:
                if isinstance(val, tuple):
                    v = val[0]
                    val = val[1:]
                else:
                    v = val
                args_seq.append((opt, v))
                if isinstance(val, tuple):
                    if len(val) == 0:
                        del ctx.params[opt]
                    else:
                        ctx.params[opt] = v
                else:
                    del ctx.params[opt]
                break
            pass
    return args_seq



import dask.array as da
import xarray as xr
from numpy import ndarray, zeros
from numpy.random import RandomState
from numpy.testing import assert_, assert_allclose
from pandas import DataFrame

from limix.qc import normalise_covariance


def test_qc_kinship_numpy():
    random = RandomState(0)
    X = random.randn(3, 5)
    K = X.dot(X.T)
    K1 = zeros((3, 3))

    K0 = normalise_covariance(K)
    K2 = normalise_covariance(K, out=K1)

    Kf = [
        [2.5990890007787586, -0.1951278087849671, 0.5472860002747189],
        [-0.1951278087849671, 0.4202620710126438, 0.2642930556468809],
        [0.5472860002747189, 0.2642930556468809, 0.5971001753452302],
    ]

    assert_allclose(K0, Kf)
    assert_allclose(K0, K1)
    assert_allclose(K0, K2)
    assert_(K2 is K1)


def test_qc_kinship_dataframe():
    random = RandomState(0)
    X = random.randn(3, 5)
    K = X.dot(X.T)
    K = DataFrame(K)

    K1 = DataFrame(zeros((3, 3)))

    K0 = normalise_covariance(K)
    K2 = normalise_covariance(K, out=K1)

    Kf = [
        [2.5990890007787586, -0.1951278087849671, 0.5472860002747189],
        [-0.1951278087849671, 0.4202620710126438, 0.2642930556468809],
        [0.5472860002747189, 0.2642930556468809, 0.5971001753452302],
    ]

    assert_allclose(K0, Kf)
    assert_(isinstance(K0, DataFrame))

    assert_allclose(K0, K1)
    assert_(isinstance(K1, DataFrame))

    assert_allclose(K0, K2)
    assert_(K2 is K1)


def test_qc_kinship_dask_array():
    random = RandomState(0)
    X = random.randn(3, 5)
    K = X.dot(X.T)
    K = da.from_array(K, chunks=2)

    K1 = zeros((3, 3))

    K0 = normalise_covariance(K)
    K2 = normalise_covariance(K, out=K1)

    Kf = [
        [2.5990890007787586, -0.1951278087849671, 0.5472860002747189],
        [-0.1951278087849671, 0.4202620710126438, 0.2642930556468809],
        [0.5472860002747189, 0.2642930556468809, 0.5971001753452302],
    ]

    assert_allclose(K0, Kf)
    assert_(isinstance(K0, da.Array))

    assert_allclose(K0, K1)
    assert_(isinstance(K1, ndarray))
    assert_(isinstance(K2, ndarray))

    assert_allclose(K0, K2)
    assert_(K2 is K1)


def test_qc_kinship_dataarray():
    random = RandomState(0)
    X = random.randn(3, 5)
    K = X.dot(X.T)
    K = da.from_array(K, chunks=2)
    K = xr.DataArray(K)

    K1 = zeros((3, 3))

    K0 = normalise_covariance(K)
    K2 = normalise_covariance(K, out=K1)

    Kf = [
        [2.5990890007787586, -0.1951278087849671, 0.5472860002747189],
        [-0.1951278087849671, 0.4202620710126438, 0.2642930556468809],
        [0.5472860002747189, 0.2642930556468809, 0.5971001753452302],
    ]

    assert_allclose(K0, Kf)
    assert_(isinstance(K0, xr.DataArray))

    assert_allclose(K0, K1)
    assert_(isinstance(K1, ndarray))
    assert_(isinstance(K2, ndarray))

    assert_allclose(K0, K2)
    assert_(K2 is K1)



def read_phenotype(filepath, verbose=True):
    """
    Read a BIMBAM phenotype file.

    Parameters
    ----------
    filepath : str
        File path.

    Returns
    -------
    :class:`pandas.DataFrame`
        DataFrame representation of the file.
    verbose : bool
        `True` for progress information; `False` otherwise.

    Examples
    --------
    .. doctest::

        >>> import limix
        >>> from limix import file_example
        >>>
        >>> with file_example("phenotype.gemma") as filepath:
        ...     print(limix.io.bimbam.read_phenotype(filepath, verbose=False))
        trait         0        1        2
        sample
        0        1.20000 -0.30000 -1.50000
        1            nan  1.50000  0.30000
        2        2.70000  1.10000      nan
        3       -0.20000 -0.70000  0.80000
        4        3.30000  2.40000  2.10000

    Notes
    -----
    BIMBAM phenotype files do not explicitly define sample ids (nor trait ids) but their
    order of appearance is used to associate samples from different files. Therefore,
    we denote the first sample found in this file as ``0``, the second as ``1``, and so
    on. We apply the same reasoning for trait naming.
    """
    from pandas import read_csv
    from .._display import session_line

    with session_line("Reading `{}`... ".format(filepath), disable=not verbose):
        df = read_csv(filepath, sep=r"\s+", header=None)

    df.index.name = "sample"
    df.columns.name = "trait"

    return df


def _see_phenotype(filepath, verbose=True):
    """
    Shows a summary of a BIMBAM phenotype file.

    Parameters
    ----------
    filepath : str
        File path.

    Returns
    -------
    str
        File representation.
    """
    from .._display import add_title_header

    df = read_phenotype(filepath, verbose)

    print(add_title_header("Phenotypes", df))



from numpy import asarray, concatenate, sort, zeros
from numpy.random import RandomState
from numpy.testing import assert_allclose
from pandas import DataFrame

from limix.stats import confusion_matrix


def test_stats_confusion():
    random = RandomState(0)

    pos1 = sort(asarray(random.randint(1, 10000, size=100), float))
    causal1 = zeros(100, dtype=bool)
    causal1[1] = True
    causal1[50] = True
    pv1 = random.rand(100)

    pos2 = sort(asarray(random.randint(1, 10000, size=25), float))
    causal2 = zeros(25, dtype=bool)
    causal2[3] = True
    pv2 = random.rand(25)

    df = DataFrame(
        data=dict(
            chrom=["1"] * 100 + ["2"] * 25,
            pv=concatenate([pv1, pv2]),
            pos=concatenate([pos1, pos2]),
            causal=concatenate([causal1, causal2]),
        )
    )

    cm = confusion_matrix(df, wsize=5)

    assert_allclose(cm.TP[90], 2)
    assert_allclose(cm.FP[90], 88)
    assert_allclose(cm.TN[90], 34)
    assert_allclose(cm.FN[90], 1)
    assert_allclose(cm.sensitivity[90], 0.666666666667)
    assert_allclose(cm.tpr[90], 0.666666666667)
    assert_allclose(cm.recall[90], 0.666666666667)
    assert_allclose(cm.specifity[90], 0.27868852459)
    assert_allclose(cm.precision[90], 0.0222222222222)
    assert_allclose(cm.npv[90], 0.971428571429)
    assert_allclose(cm.fallout[90], 0.72131147541)
    assert_allclose(cm.fpr[90], 0.72131147541)
    assert_allclose(cm.fnr[90], 0.333333333333)
    assert_allclose(cm.fdr[90], 0.977777777778)
    assert_allclose(cm.accuracy[90], 0.288)
    assert_allclose(cm.f1score[90], 0.0430107526882)
    (fpr, tpr) = cm.roc()
    assert_allclose(fpr[90], 0.729508196721)
    assert_allclose(tpr[90], 0.666666666667)



from .._display import session_line


def extract(filepath, verbose=True):
    """
    Extract a compressed file.

    Parameters
    ----------
    filepath : str
        File path.
    verbose : bool, optional
        ``True`` for displaying progress. Defaults to ``True``.
    """
    formats_order = _best_order(filepath)

    err_msgs = []
    with session_line("Extracting {}... ".format(filepath), disable=not verbose):

        for f in formats_order:
            try:
                return f(filepath)
            except Exception as e:
                err_msgs.append(str(e))

        raise RuntimeError(f"Could not extract `{filepath}`.\n" + "\n".join(err_msgs))


def _extract_tar(filepath):
    import tarfile

    tar = tarfile.open(filepath)
    tar.extractall()
    filepath = tar.getnames()[0]
    tar.close()
    return filepath


def _extract_bz2(filepath):
    import bz2
    import os

    filename = os.path.splitext(filepath)[0]

    with open(filepath, "rb") as f:
        o = bz2.decompress(f.read())

    with open(filename, "wb") as f:
        f.write(o)
    return filename


def _extract_zip(filepath, dest="."):
    import zipfile

    with zipfile.ZipFile(filepath, "r") as zip_ref:
        zip_ref.extractall(dest)

    return dest


def _best_order(filepath):
    formats = [_extract_tar, _extract_bz2, _extract_zip]
    formats_order = []
    if filepath.endswith(".zip"):
        formats_order = [_extract_zip]
    elif filepath.endswith(".tar.gz"):
        formats_order = [_extract_tar]
    elif filepath.endswith(".bz2"):
        formats_order = [_extract_bz2]

    for f in formats:
        if f not in formats_order:
            formats_order.append(f)

    return formats_order



import pytest
import scipy.stats as st
from numpy import (
    argmin,
    array,
    concatenate,
    dot,
    exp,
    eye,
    kron,
    nan,
    reshape,
    sqrt,
    zeros,
)
from numpy.random import RandomState
from numpy.testing import assert_allclose
from pandas import DataFrame

from limix.qc import normalise_covariance
from limix.qtl import scan
from limix.stats import linear_kinship, multivariate_normal as mvn


def _test_qtl_scan_st(lik):
    random = RandomState(0)
    n = 30
    ncovariates = 3

    M = random.randn(n, ncovariates)

    v0 = random.rand()
    v1 = random.rand()

    G = random.randn(n, 4)

    K = random.randn(n, n + 1)
    K = normalise_covariance(K @ K.T)

    beta = random.randn(ncovariates)
    alpha = random.randn(G.shape[1])

    m = M @ beta + G @ alpha
    y = mvn(random, m, v0 * K + v1 * eye(n))

    idx = [[0, 1], 2, [3]]

    if lik == "poisson":
        y = random.poisson(exp(y))
    elif lik == "bernoulli":
        y = random.binomial(1, 1 / (1 + exp(-y)))
    elif lik == "probit":
        y = random.binomial(1, st.norm.cdf(y))
    elif lik == "binomial":
        ntrials = random.randint(0, 30, len(y))
        y = random.binomial(ntrials, 1 / (1 + exp(-y)))
        lik = (lik, ntrials)

    r = scan(G, y, lik=lik, idx=idx, K=K, M=M, verbose=False)
    str(r)
    str(r.stats.head())
    str(r.effsizes["h2"].head())
    str(r.h0.trait)
    str(r.h0.likelihood)
    str(r.h0.lml)
    str(r.h0.effsizes)
    str(r.h0.variances)


def test_qtl_scan_st():
    _test_qtl_scan_st("normal")
    _test_qtl_scan_st("poisson")
    _test_qtl_scan_st("bernoulli")
    _test_qtl_scan_st("probit")
    _test_qtl_scan_st("binomial")


def test_qtl_scan_three_hypotheses_mt():
    random = RandomState(0)
    n = 30
    ntraits = 2
    ncovariates = 3

    A = random.randn(ntraits, ntraits)
    A = A @ A.T
    M = random.randn(n, ncovariates)

    C0 = random.randn(ntraits, ntraits)
    C0 = C0 @ C0.T

    C1 = random.randn(ntraits, ntraits)
    C1 = C1 @ C1.T

    G = random.randn(n, 4)

    A0 = random.randn(ntraits, 1)
    A1 = random.randn(ntraits, 2)
    A01 = concatenate((A0, A1), axis=1)

    K = random.randn(n, n + 1)
    K = normalise_covariance(K @ K.T)

    beta = vec(random.randn(ntraits, ncovariates))
    alpha = vec(random.randn(A01.shape[1], G.shape[1]))

    m = kron(A, M) @ beta + kron(A01, G) @ alpha
    Y = unvec(mvn(random, m, kron(C0, K) + kron(C1, eye(n))), (n, -1))

    idx = [[0, 1], 2, [3]]
    r = scan(G, Y, idx=idx, K=K, M=M, A=A, A0=A0, A1=A1, verbose=False)
    str(r)


def test_qtl_scan_two_hypotheses_mt():
    random = RandomState(0)
    n = 30
    ntraits = 2
    ncovariates = 3

    A = random.randn(ntraits, ntraits)
    A = A @ A.T
    M = random.randn(n, ncovariates)

    C0 = random.randn(ntraits, ntraits)
    C0 = C0 @ C0.T

    C1 = random.randn(ntraits, ntraits)
    C1 = C1 @ C1.T

    G = random.randn(n, 4)

    A0 = random.randn(ntraits, 1)
    A1 = random.randn(ntraits, 2)
    A01 = concatenate((A0, A1), axis=1)

    K = random.randn(n, n + 1)
    K = normalise_covariance(K @ K.T)

    beta = vec(random.randn(ntraits, ncovariates))
    alpha = vec(random.randn(A01.shape[1], G.shape[1]))

    m = kron(A, M) @ beta + kron(A01, G) @ alpha
    Y = unvec(mvn(random, m, kron(C0, K) + kron(C1, eye(n))), (n, -1))

    idx = [[0, 1], 2, [3]]
    r = scan(G, Y, idx=idx, K=K, M=M, A=A, A1=A1, verbose=False)
    str(r)


def test_qtl_scan_lmm():
    random = RandomState(0)
    nsamples = 50

    G = random.randn(50, 100)
    K = linear_kinship(G[:, 0:80], verbose=False)

    y = dot(G, random.randn(100)) / sqrt(100) + 0.2 * random.randn(nsamples)

    M = G[:, :5]
    X = G[:, 68:70]

    result = scan(X, y, lik="normal", K=K, M=M, verbose=False)

    pv = result.stats["pv20"]

    ix_best_snp = argmin(array(pv))
    M = concatenate((M, X[:, [ix_best_snp]]), axis=1)
    result = scan(X, y, "normal", K, M=M, verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(pv[ix_best_snp], 1.0, atol=1e-6)


def test_qtl_scan_lmm_nokinship():
    random = RandomState(0)
    nsamples = 50

    G = random.randn(50, 100)
    K = linear_kinship(G[:, 0:80], verbose=False)

    y = dot(G, random.randn(100)) / sqrt(100) + 0.2 * random.randn(nsamples)

    M = G[:, :5]
    X = G[:, 68:70]

    result = scan(X, y, "normal", K, M=M, verbose=False)
    pv = result.stats["pv20"].values
    assert_allclose(pv[:2], [8.159539103135342e-05, 0.10807353641893498], atol=1e-5)


def test_qtl_scan_lmm_repeat_samples_by_index():
    random = RandomState(0)
    nsamples = 30
    samples = ["sample{}".format(i) for i in range(nsamples)]

    G = random.randn(nsamples, 100)
    G = DataFrame(data=G, index=samples)

    K = linear_kinship(G.values[:, 0:80], verbose=False)
    K = DataFrame(data=K, index=samples, columns=samples)

    y0 = dot(G, random.randn(100)) / sqrt(100) + 0.2 * random.randn(nsamples)
    y1 = dot(G, random.randn(100)) / sqrt(100) + 0.2 * random.randn(nsamples)
    y = concatenate((y0, y1))
    y = DataFrame(data=y, index=samples + samples)

    M = G.values[:, :5]
    X = G.values[:, 68:70]
    M = DataFrame(data=M, index=samples)
    X = DataFrame(data=X, index=samples)

    result = scan(X, y, "normal", K, M=M, verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(pv.values[0], 0.9920306566395604, rtol=1e-6)

    ix_best_snp = argmin(array(result.stats["pv20"]))

    M = concatenate((M, X.loc[:, [ix_best_snp]]), axis=1)
    M = DataFrame(data=M, index=samples)

    result = scan(X, y, "normal", K, M=M, verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(pv[ix_best_snp], 1.0, rtol=1e-6)
    assert_allclose(pv.values[0], 0.6684700834450028, rtol=1e-6)


def test_qtl_scan_glmm_binomial():
    random = RandomState(0)
    nsamples = 25

    X = random.randn(nsamples, 2)
    G = random.randn(nsamples, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 100, nsamples)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i, nt in enumerate(ntrials):
        for _ in range(nt):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    result = scan(X, successes, ("binomial", ntrials), K, verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(pv, [0.9315770010211236, 0.8457015828837173], atol=1e-6, rtol=1e-6)


def test_qtl_scan_glmm_wrong_dimensions():
    random = RandomState(0)
    nsamples = 25

    X = random.randn(nsamples, 2)
    G = random.randn(nsamples, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 100, nsamples)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i, nt in enumerate(ntrials):
        for _ in range(nt):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    M = random.randn(49, 2)
    scan(X, successes, ("binomial", ntrials), K, M=M, verbose=False)


def test_qtl_scan_glmm_bernoulli():
    random = RandomState(0)
    nsamples = 25

    X = random.randn(nsamples, 2)
    G = random.randn(nsamples, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 2, nsamples)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i, nt in enumerate(ntrials):
        for _ in range(nt):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    result = scan(X, successes, "bernoulli", K, verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(pv, [0.3399326545917558, 0.8269454251659921], rtol=1e-5)


def test_qtl_scan_glmm_bernoulli_nokinship():
    random = RandomState(0)
    nsamples = 25

    X = random.randn(nsamples, 2)
    G = random.randn(nsamples, 100)
    ntrials = random.randint(1, 2, nsamples)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i, nt in enumerate(ntrials):
        for _ in range(nt):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    result = scan(X, successes, "bernoulli", verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(pv, [0.3399067917883736, 0.8269568797830423], rtol=1e-5)


def test_qtl_scan_lm():
    random = RandomState(0)
    nsamples = 25

    G = random.randn(nsamples, 100)

    y = dot(G, random.randn(100)) / sqrt(100) + 0.2 * random.randn(nsamples)

    M = G[:, :5]
    X = G[:, 5:]
    result = scan(X, y, "normal", M=M, verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(pv[:2], [0.02625506841465465, 0.9162689001409643], rtol=1e-5)


def test_qtl_scan_gmm_binomial():
    random = RandomState(0)
    nsamples = 25

    X = random.randn(nsamples, 2)
    ntrials = random.randint(1, nsamples, nsamples)
    z = dot(X, random.randn(2))

    successes = zeros(len(ntrials), int)
    for i in range(len(ntrials)):
        for _ in range(ntrials[i]):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    result = scan(X, successes, ("binomial", ntrials), verbose=False)
    pv = result.stats["pv20"]
    assert_allclose(
        pv, [2.4604711379400065e-06, 0.01823278752006871], rtol=1e-5, atol=1e-5
    )


def test_qtl_finite():
    random = RandomState(0)
    nsamples = 20

    X = random.randn(50, 2)
    G = random.randn(50, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 100, nsamples)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i, nt in enumerate(ntrials):
        for _ in range(nt):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    successes = successes.astype(float)
    ntrials = ntrials.astype(float)

    successes[0] = nan
    with pytest.raises(ValueError):
        scan(X, successes, ("binomial", ntrials), K, verbose=False)
    successes[0] = 1.0

    K[0, 0] = nan
    with pytest.raises(ValueError):
        scan(X, successes, ("binomial", ntrials), K, verbose=False)
    K[0, 0] = 1.0

    X[0, 0] = nan
    with pytest.raises(ValueError):
        scan(X, successes, ("binomial", ntrials), K, verbose=False)
    X[0, 0] = 1.0


def vec(x):
    return reshape(x, (-1,) + x.shape[2:], order="F")


def unvec(x, shape):
    return reshape(x, shape, order="F")



from ._ifactory import IScanResultFactory
from ._mt_factory import MTScanResultFactory
from ._st_factory import STScanResultFactory

__all__ = ["STScanResultFactory", "MTScanResultFactory", "IScanResultFactory"]



import os

from numpy import nan
from numpy.random import RandomState
from numpy.testing import assert_allclose

import limix
from limix._cli.pipeline import Pipeline
from limix._cli.preprocess import impute, normalize, where
from limix.io._fetch import fetch


def test_pipeline_where_filter():

    filenames = [
        "chrom22_subsample20_maf0.10.bed",
        "chrom22_subsample20_maf0.10.fam",
        "chrom22_subsample20_maf0.10.bim",
    ]
    shapes = [
        (274, 5647),
        (274, 0),
        (274, 49008),
        (274, 49008),
        (4, 49008),
        (274, 5647),
        (274, 0),
        (274, 49008),
    ]
    specs = [
        "genotype:(16050612 <= pos) & (pos < 21050612)",
        "genotype:(16050612 <= pos) & (pos < 16050612)",
        "genotype",
        "genotype:",
        "genotype:sample.isin(['HG00111', 'HG00112', 'NA20775', 'NA20804'])",
        "genotype:(chrom == '22') & (16050612 <= pos) & (pos < 21050612)",
        "genotype: a0 == a1",
        "genotype: sample != None",
    ]
    random = RandomState(0)
    with limix.file_example(filenames) as filepaths:
        folder = os.path.dirname(filepaths[0])
        filepath = os.path.join(folder, "chrom22_subsample20_maf0.10")

        for shape, spec in zip(shapes, specs):
            G = fetch("genotype", f"{filepath}", verbose=False)
            y = random.randn(G.shape[0])
            data = {"G": G, "y": y}
            pipeline = Pipeline(data)
            pipeline.append(where, "where", spec)
            data = pipeline.run(verbose=False)
            assert data["G"].shape == shape
            assert data["G"].dims == ("sample", "candidate")


def test_pipeline_normalize():

    with limix.file_example("expr.csv") as filepath:
        folder = os.path.dirname(filepath)
        filepath = os.path.join(folder, "expr.csv")

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(normalize, "normalize", spec="trait:trait:gaussianize")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [
                -2.684_091_161_327_623_7,
                -0.348_755_695_517_044_7,
                -1.429_964_275_028_744_2,
            ],
        )

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(normalize, "normalize", spec="trait:sample:gaussianize")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [-1.382_994_127_100_638, 0.0, -0.674_489_750_196_081_7],
        )


def test_pipeline_impute():

    with limix.file_example("expr_nan.csv") as filepath:
        folder = os.path.dirname(filepath)
        filepath = os.path.join(folder, "expr_nan.csv")

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(impute, "impute", spec="trait:trait:mean")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [0.086_810_659_809_597_32, -0.351_453_183_202_999_95, -1.319_967_938_81],
        )


def test_pipeline_normalize_nan():
    with limix.file_example("expr_nan.csv") as filepath:
        folder = os.path.dirname(filepath)
        filepath = os.path.join(folder, "expr_nan.csv")

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(normalize, "normalize", spec="trait:sample:gaussianize")
        data = pipeline.run(verbose=False)
        assert_allclose(data["y"].values[0, :3], [nan, 0.0, -0.841_621_233_572_914_3])

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(normalize, "normalize", spec="trait:trait:gaussianize")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [nan, -0.345_222_629_722_377_3, -1.429_964_275_028_744_2],
        )

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(normalize, "normalize", spec="trait::gaussianize")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [nan, -0.345_222_629_722_377_3, -1.429_964_275_028_744_2],
        )

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(normalize, "normalize", spec="trait")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [nan, -0.345_222_629_722_377_3, -1.429_964_275_028_744_2],
        )


def test_pipeline_impute_and_normalize():

    with limix.file_example("expr_nan.csv") as filepath:
        folder = os.path.dirname(filepath)
        filepath = os.path.join(folder, "expr_nan.csv")

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(impute, "impute", spec="trait:trait:mean")
        pipeline.append(normalize, "normalize", spec="trait:trait:gaussianize")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [
                -0.013_672_943_874_439_595,
                -0.348_755_695_517_044_7,
                -1.429_964_275_028_744_2,
            ],
        )

        y = fetch("trait", f"{filepath}::row=trait", verbose=False)
        data = {"y": y}
        pipeline = Pipeline(data)
        pipeline.append(impute, "impute", spec="trait::")
        pipeline.append(normalize, "normalize", spec="trait:")
        data = pipeline.run(verbose=False)
        assert_allclose(
            data["y"].values[0, :3],
            [
                -0.013_672_943_874_439_595,
                -0.348_755_695_517_044_7,
                -1.429_964_275_028_744_2,
            ],
        )



from numpy.random import RandomState
from numpy.testing import assert_allclose

from limix.stats import empirical_pvalues, multipletests


def test_empirical_pvalues():

    random = RandomState(1)

    x0 = random.chisquare(1, 1000)
    x1 = random.chisquare(1, 10000)

    pv = empirical_pvalues(x0, x1)

    assert_allclose(pv[:3], [0.5599, 1.0, 0.8389])
    assert_allclose(pv[-3:], [0.249, 0.3278, 0.4848])


def test_multipletests():
    random = RandomState(5)

    pv = multipletests(random.chisquare(1, 10))[1]
    assert_allclose(
        pv,
        [
            0.56399935,
            0.55249248,
            0.96090415,
            0.99606969,
            0.74247325,
            0.12299317,
            1.0,
            0.91970786,
            0.99606969,
            0.99595594,
        ],
    )



def read(prefix, verbose=True):
    """
    Read PLINK files into Pandas data frames.

    Parameters
    ----------
    prefix : str
        Path prefix to the set of PLINK files.
    verbose : bool
        ``True`` for progress information; ``False`` otherwise.

    Returns
    -------
    alleles : pandas dataframe
    samples : pandas dataframe
    genotype : ndarray

    Examples
    --------
    .. doctest::

        >>> from limix.io import plink
        >>> from pandas_plink import example_file_prefix
        >>>
        >>> (bim, fam, bed) = plink.read(example_file_prefix(), verbose=False)
        >>> print(bim.head())
                   chrom         snp       cm    pos a0 a1  i
        candidate
        rs10399749     1  rs10399749  0.00000  45162  G  C  0
        rs2949420      1   rs2949420  0.00000  45257  C  T  1
        rs2949421      1   rs2949421  0.00000  45413  0  0  2
        rs2691310      1   rs2691310  0.00000  46844  A  T  3
        rs4030303      1   rs4030303  0.00000  72434  0  G  4
        >>> print(fam.head())
                       fid       iid    father    mother gender trait  i
        sample
        Sample_1  Sample_1  Sample_1         0         0      1    -9  0
        Sample_2  Sample_2  Sample_2         0         0      2    -9  1
        Sample_3  Sample_3  Sample_3  Sample_1  Sample_2      2    -9  2
        >>> print(bed.compute())
        [[ 2.  2.  1.]
         [ 2.  1.  2.]
         [nan nan nan]
         [nan nan  1.]
         [ 2.  2.  2.]
         [ 2.  2.  2.]
         [ 2.  1.  0.]
         [ 2.  2.  2.]
         [ 1.  2.  2.]
         [ 2.  1.  2.]]

    Notice the ``i`` column in bim and fam data frames. It maps to the
    corresponding position of the bed matrix:

    .. doctest::

        >>> from limix.io import plink
        >>> from pandas_plink import example_file_prefix
        >>>
        >>> (bim, fam, bed) = plink.read(example_file_prefix(), verbose=False)
        >>> chrom1 = bim.query("chrom=='1'")
        >>> X = bed[chrom1.i.values, :].compute()
        >>> print(X)
        [[ 2.  2.  1.]
         [ 2.  1.  2.]
         [nan nan nan]
         [nan nan  1.]
         [ 2.  2.  2.]
         [ 2.  2.  2.]
         [ 2.  1.  0.]
         [ 2.  2.  2.]
         [ 1.  2.  2.]
         [ 2.  1.  2.]]
    """
    from pandas_plink import read_plink
    from .._display import session_line

    with session_line("Reading `{}`...\n".format(prefix), disable=not verbose):
        data = read_plink(prefix, verbose=verbose)
        if verbose:
            # Clear up the progress bar and get back to the initial line.
            print("\033[1A\033[K\033[1A", end="")

        data[1].name = "fam"
        data[1].index = data[1]["iid"]
        data[1].index.name = "sample"

        data[0].name = "bim"
        data[0].index = data[0]["snp"].astype(str).values
        data[0].index.name = "candidate"

    return data


def _read_dosage(prefix, verbose):
    from pandas_plink import read_plink

    return read_plink(prefix, verbose=verbose)[2].T


def _see_bed(filepath, verbose):
    from .._display import add_title_header

    (bim, fam, _) = read(filepath, verbose=verbose)

    print(add_title_header("Samples", bim))
    print(add_title_header("Genotype", fam))


def _see_kinship(filepath, verbose):
    from .. import plot
    from .._display import session_line

    if filepath.endswith(".grm.raw"):
        with session_line("Reading {}... ".format(filepath), disable=not verbose):
            K = _read_grm_raw(filepath)
    else:
        print("File %s not found." % filepath)
        return

    plot.kinship(K)


def _read_grm_raw(filepath):
    from numpy import loadtxt

    return loadtxt(filepath)



from numpy.random import RandomState
from numpy.testing import assert_allclose

from limix.stats import Chi2Mixture


def test_chi2mixture():
    dof = 2
    mixture = 0.2
    n = 100

    random = RandomState(1)
    x = random.chisquare(dof, n)
    n0 = int((1 - mixture) * n)
    idxs = random.choice(n, n0, replace=False)
    x[idxs] = 0

    chi2mix = Chi2Mixture(
        scale_min=0.1,
        scale_max=5.0,
        dof_min=0.1,
        dof_max=5.0,
        qmax=0.1,
        tol=4e-3,
        lrt=x,
    )
    chi2mix.estimate_chi2mixture(x)
    pv = chi2mix.sf([0.0, 0.2])
    assert_allclose(pv, [0.19999999999999996, 0.1412935752078675])
    assert_allclose(chi2mix.scale, 1.9808080808080812)
    assert_allclose(chi2mix.dof, 0.891919191919192)
    assert_allclose(chi2mix.mixture, 0.199999999999999960)
